from __future__ import print_function

import numpy as np

from ..utils.gpu import cp
from ..utils     import cr_nvtx as cr, raiseError
from ..vmmath    import temporal_mean, subtract_mean, matmul, vector_sum, least_squares, ridge_regresion
from ..POD       import run, truncate


class GappyPOD:
	def __init__(
		self,
		centered=False,
		apply_truncation=False,
		truncation_param=-0.99,
		reconstruction_method="standard",
		ridge_lambda=0.01,
	):
		"""
		Gappy POD model for reconstructing incomplete data using POD modes.

		Args:
			centered (bool): Whether to center the data by subtracting the mean.
			apply_truncation (bool): Whether to apply truncation.
			truncation_method (float or int): Threshold for truncation.
			reconstruction_method (str): Reconstruction method ('standard' or 'ridge').
			ridge_lambda (float): Regularization parameter for ridge reconstruction.
		"""
		# Validate reconstruction method
		if not reconstruction_method.lower() in ["standard", "ridge"]:
			raiseError("Reconstruction method must be either 'standard' or 'ridge'.")

		self.centered            = centered
		self.truncate            = apply_truncation
		self.truncation_param    = truncation_param
		self.reconstruction_type = least_squares if reconstruction_method.lower() == "standard" else lambda a,b : ridge_regresion(a,b,ridge_lambda)

		# Attributes to store fitted data
		self.mean               = None
		self.U_truncated_scaled = None

	@cr('GPOD.fit')
	def fit(self, snapshot_matrix: np.ndarray, **kwargs) -> None:
		"""
		Fit the Gappy POD model using the snapshot matrix.

		Args:
			snapshot_matrix  (np.ndarray): Training matrix [n_features, n_samples].
		"""
		cnp = cp if type(snapshot_matrix) is cp.ndarray else np
		# Center data if required
		self.mean = temporal_mean(snapshot_matrix) if self.centered else cnp.zeros(snapshot_matrix.shape[0])
		X = subtract_mean(snapshot_matrix,self.mean)

		# Compute SVD through POD this way we can reuse some nice features
		self.U_truncated, self.S_truncated, VT = run(X, remove_mean=False, **kwargs)

		# Apply truncation if specified
		if self.truncate:
			self.U_truncated, self.S_truncated, _ = truncate(self.U_truncated,self.S_truncated,VT,self.truncation_param)

		# Masked and scaled POD modes
		self.U_truncated_scaled = self.U_truncated * self.S_truncated

	@cr('GPOD.predict')
	def predict(self, gappy_vector: np.ndarray) -> np.ndarray:
		"""
		Reconstruct missing data using the fitted Gappy POD model.

		Args:
			gappy_vector (np.ndarray): Sparse vector with missing values.

		Returns:
			np.ndarray: Reconstructed data vector.
		"""
		if self.U_truncated_scaled is None:
			raiseError("The model must be fitted before calling predict.")

		# Prepare the masked input
		mask        = (gappy_vector != 0) # Binary mask for observed data
		gappy_input = gappy_vector - self.mean*mask
		PT_U        = mask[:,None]*self.U_truncated_scaled

		# Solve for coefficients
		coef = self.reconstruction_type(PT_U,gappy_input)

		# Reconstruct missing data
		vector_reconstructed = matmul(self.U_truncated_scaled,coef[:,None]).flatten() + self.mean
		return vector_reconstructed

	@cr('GPOD.reconstruct')
	def reconstruct_full_set(
		self, incomplete_snapshot: np.ndarray, iter_num: int
	) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
		"""
		Iteratively reconstruct an incomplete snapshot matrix using Gappy POD.

		Args:
			incomplete_snapshot (np.ndarray): Snapshot matrix with missing values (n_features, n_samples).
			iter_num (int): Number of iterations for the iterative reconstruction process.

		Returns:
			tuple: Reconstructed snapshot matrix, eigenvalue spectrum across iterations,
				   and cumulative energy.
		"""
		cnp = cp if type(incomplete_snapshot) is cp.ndarray else np
		# Step 1: Create mask for missing values
		mask_incomplete = (incomplete_snapshot != 0).astype(int)

		# Step 2: Compute row-wise mean for non-missing values
		g_i_mean = cnp.array([vector_sum(incomplete_snapshot[i, :]) / vector_sum(mask_incomplete[i, :].astype(incomplete_snapshot.dtype))
		for i in range(incomplete_snapshot.shape[0])])

		# Initialize the reconstructed snapshot matrix
		h_recons = incomplete_snapshot.copy()
		for i in range(h_recons.shape[0]):
			h_recons[i, mask_incomplete[i] == 0] = g_i_mean[i]

		# Step 3: Initialize arrays to store results
		eig_spec_iter = cnp.empty((h_recons.shape[1], iter_num))
		c_e = cnp.empty((h_recons.shape[1], iter_num))

		# Iterative reconstruction
		for k in range(iter_num):
			# Fit the model with the current reconstructed snapshot matrix
			self.fit(h_recons)

			# Reconstruct all snapshots
			snapshots_recons = cnp.empty(h_recons.shape)
			for i in range(h_recons.shape[1]):
				gappy_vector = cnp.copy(incomplete_snapshot[:, i])
				snapshots_recons[:, i] = self.predict(gappy_vector)

			# Update the reconstruction matrix
			h_recons = incomplete_snapshot.copy()
			for j in range(h_recons.shape[1]):
				mask = mask_incomplete[:, j] == 0
				h_recons[mask, j] = snapshots_recons[mask, j]

			# Compute the eigenvalue spectrum and cumulative energy
			_, S, _ = run(h_recons, remove_mean=True)
			eig_spec_iter[:, k] = S**2/ vector_sum(S**2,0)
			normS = vector_sum(S,0)
			cumulative = 0
			for ii in range(S.shape[0]):
				cumulative += S[ii]
				c_e[ii, k] = cumulative / normS
				
		return h_recons, eig_spec_iter, c_e



#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - SVD.
#
# Last rev: 27/10/2021
from __future__ import print_function, division

import time, numpy as np

from ..utils.gpu import cp
from .maths      import matmul, matmulp
from ..utils     import cr_nvtx as cr, MPI_RANK, MPI_SIZE, mpi_send, mpi_recv


def next_power_of_2(n):
	'''
	Find the next power of 2 of n
	'''
	p = 1
	if (n and not(n & (n - 1))):
		return n
	while (p < n): p <<= 1
	return p


@cr('math.qr')
def qr(A):
	'''
	QR factorization using Lapack
		Q(m,n) is the Q matrix
		R(n,n) is the R matrix
	'''
	p = cp if type(A) is cp.ndarray else np
	return p.linalg.qr(A)

@cr('math.svd')
def svd(A,method='gesdd'):
	'''
	Single value decomposition (SVD) using numpy.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	p = cp if type(A) is cp.ndarray else np
	return p.linalg.svd(A,full_matrices=False)

@cr('math.tsqr')
def tsqr(Ai):
	'''
	Parallel QR factorization of a real array using Lapack
		Q(m,n) is the Q matrix
		R(n,n) is the R matrix
	'''
	p = cp if type(Ai) is cp.ndarray else np
	_, n = Ai.shape
	# Algorithm 1 from Demmel et al (2012)
	# 1: QR Factorization on Ai to obtain Q1i and Ri
	Q1i, R    = qr(Ai)
	nextPower = next_power_of_2(MPI_SIZE)
	nlevels   = int(np.log2(nextPower))
	QW        = p.eye(n, dtype = Ai.dtype)
	C         = p.zeros((2*n, n), Ai.dtype)
	Q2l       = p.zeros((2*n*nlevels, n), Ai.dtype)
	blevel    = 1
	for ilevel in range(nlevels):
		# Store R in the upper part of the C matrix
		C[:n, :] = R
		# Decide who sends and who recieves, use R as buffer
		prank = MPI_RANK ^ blevel
		if MPI_RANK & blevel:
			if prank < MPI_SIZE:
				mpi_send(R, prank)
		else:
			if prank < MPI_SIZE:
				R = mpi_recv(source = prank)
				# Store R in the lower part of the C matrix
				C[n:, :] = R
				# 2: QR from the C matrix, reuse C and R
				Q2i, R = qr(C)
				# Store Q2i from this level
				Q2l[2*n*ilevel:2*n*ilevel+2*n, :] = Q2i
		blevel <<= 1
	# At this point R is correct on processor 0
	# Broadcast R and its part of the Q matrix
	if MPI_SIZE > 1:
		blevel = 1 << (nlevels - 1)
		mask   = blevel - 1
	for ilevel in reversed(range(nlevels)):
		if MPI_RANK & mask == 0:
			# Obtain Q2i for this level - use C as buffer
			C = Q2l[2*n*ilevel:2*n*ilevel+2*n, :]
			# Multiply by QW either set to identity or allocated to a value
			# Store into Q2i
			Q2i = matmul(C, QW)
			# Communications scheme
			prank = MPI_RANK^blevel
			if MPI_RANK & blevel:
				if prank < MPI_SIZE:
					C = mpi_recv(source = prank)
					# Recover R from the upper part of C and QW from the lower part
					R  = C[:n, :]
					QW = C[n:, :]
			else:
				if prank < MPI_SIZE:
					# Set up C matrix for sending
					# Store R in the upper part and Q2i on the lower part
					# Store Q2i of this rank to QW
					C[:n, :] = R
					C[n:, :] = Q2i[n:, :]
					QW       = Q2i[:n, :]
					mpi_send(C, prank)
		blevel >>= 1
		mask   >>= 1
	# Multiply Q1i and QW to obtain Qi
	Qi = matmul(Q1i, QW)
	return Qi,R

@cr('math.randomized_qr')
def randomized_qr(Ai, r, q, seed=-1):
	'''
	Ai(m,n)  data matrix dispersed on each processor.
	r        target number of modes

	Qi(m,r)  
	B (r,n) 
	'''
	p = cp if type(Ai) is cp.ndarray else np
	_, n  = Ai.shape
	seed = int(time.time()) if seed < 0 else seed
	p.random.seed(seed=seed)
	omega = p.random.rand(n, r).astype(Ai.dtype)
	Yi = matmul(Ai,omega)
	# QR factorization on A
	for j in range(q):
		Qi,_ = tsqr(Yi)
		Q2i  = matmulp(Ai.T,Qi)
		Yi   = matmul(Ai,Q2i)

	Qi,_ = tsqr(Yi)
	B    = matmulp(Qi.T,Ai)

	return Qi, B

@cr('math.init_qr_streaming')
def init_qr_streaming(Ai, r, q, seed=None):
	'''
	Ai(m,n)  data matrix dispersed on each processor.
	r        target number of modes

	Qi(m,r)  
	B (r,n) 
	'''
	p = cp if type(Ai) is cp.ndarray else np
	_, n  = Ai.shape
	seed = int(time.time()) if seed == None else seed
	p.random.seed(seed=seed)
	omega = p.random.rand(n, r).astype(Ai.dtype)
	Yi    = matmul(Ai,omega)
	# QR factorization on A
	for j in range(q):
		Qi,_ = tsqr(Yi)
		Q2i  = matmulp(Ai.T,Qi)
		Yi   = matmul(Ai,Q2i)

	Qi,_ = tsqr(Yi)
	B    = matmulp(Qi.T,Ai)

	return Qi, B, Yi

@cr('math.qr_iteration')
def update_qr_streaming(Ai, Q1, B1, Yo, r, q):
	'''
	Ai(m,n)  data matrix dispersed on each processor.
	r        target number of modes

	Qi(m,r)  
	B (r,n) 
	'''
	p = cp if type(Ai) is cp.ndarray else np
	_, n  = Ai.shape
	omega = np.random.rand(n, r).astype(Ai.dtype)
	Yn    = matmul(Ai,omega)
	for jj in range(q):
		Qpi,_ = tsqr(Yn)
		O2    = matmulp(Ai.T, Qpi)
		Yn    = matmul(Ai,O2)
	Yo   += Yn
	Q2,_  = tsqr(Yo)
	Q2Q1  = matmulp(Q2.T, Q1)
	B2o   = matmul(Q2Q1, B1)
	B2n   = matmulp (Q2.T, Ai)
	B2    = p.hstack((B2o, B2n))

	return Q2, B2, Yo

@cr('math.tsqr_svd')
def tsqr_svd(Ai):
	'''
	Single value decomposition (SVD) using TSQR algorithm from
	J. Demmel, L. Grigori, M. Hoemmen, and J. Langou, ‘Communication-optimal Parallel
	and Sequential QR and LU Factorizations’, SIAM J. Sci. Comput.,
	vol. 34, no. 1, pp. A206–A239, Jan. 2012,

	doi: 10.1137/080731992.

	Ai(m,n)  data matrix dispersed on each processor.

	Ui(m,n)  POD modes dispersed on each processor (must come preallocated).
	S(n)     singular values.
	VT(n,n)  right singular vectors (transposed).
	'''
	# QR factorization on A
	Qi,R = tsqr(Ai)

	# Algorithm 2 from Sayadi and Schmid (2016) - Ui, S and VT
	# At this point we have R and Qi scattered on the processors
	# Call SVD routine
	Ur, S, V = svd(R)
	# Compute Ui = Qi x Ur
	Ui = matmul(Qi, Ur)
	return Ui, S, V

@cr('math.randomized_svd')
def randomized_svd(Ai, r, q, seed=-1):
	'''
	Ai(m,n)  data matrix dispersed on each processor.
	r        target number of modes

	Ui(m,n)  POD modes dispersed on each processor (must come preallocated).
	S(n)     singular values.
	VT(n,n)  right singular vectors (transposed).
	'''
	seed = int(time.time()) if seed < 0 else seed
	np.random.seed(seed=seed)

	Qi, B    = randomized_qr(Ai,r,q,seed=seed)
	Ur, S, V = svd(B)
	
	# Compute Ui = Qi x Ur
	Ui = matmul(Qi, Ur)

	return Ui, S, V


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Autoencoder architecture for NN Module
#
# Last rev: 09/10/2024

import torch
import torch.nn            as nn
import torch.nn.functional as F
import numpy               as np

from   torch.utils.data        import DataLoader
from   torch.amp               import GradScaler, autocast
from   torch.utils.tensorboard import SummaryWriter
from   torchsummary            import summary

from   functools               import reduce
from   operator                import mul

from   ..                      import DEVICE
from   ...utils                import cr, pprint


## Wrapper of a variational autoencoder
class Autoencoder(nn.Module):
    r"""
    Autoencoder class for neural network module. The model is based on the PyTorch.

    Args:
        latent_dim (int): Dimension of the latent space.
        in_shape (tuple): Shape of the input data.
        input_channels (int): Number of input channels.
        encoder (torch.nn.Module): Encoder model.
        decoder (torch.nn.Module): Decoder model.
        device (str): Device to run the model. Default is 'cuda' if available, otherwise 'cpu'.
    
    """

    def __init__(
        self,
        latent_dim: int,
        in_shape: tuple,
        input_channels: int,
        encoder: nn.Module,
        decoder: nn.Module,
        device: torch.device = DEVICE,
    ):
        super(Autoencoder, self).__init__()
        self.lat_dim  = latent_dim
        self.in_shape = in_shape
        self.inp_chan = input_channels
        self.N        = reduce(mul, in_shape)
        self.encoder  = encoder
        self.decoder  = decoder
        self._device  = device
        encoder.to(self._device)
        decoder.to(self._device)
        self.to(self._device)
        summary(self, input_size=(self.inp_chan, *self.in_shape),device=device)
      
    def _lossfunc(self, x, recon_x, reduction):
        return  F.mse_loss(recon_x.view(-1, self.N), x.view(-1, self.N),reduction=reduction)
    
    def forward(self, x):
        z     = self.encoder(x)
        recon = self.decoder(z)
        return recon, z

    def fit(
        self,
        train_dataset: torch.utils.data.Dataset,
        eval_dataset: torch.utils.data.Dataset = None,
        epochs: int = 100,
        callback=None,
        lr: float = 1e-3,
        BASEDIR: str = "./",
        reduction: str = "mean",
        lr_decay: float = 0.999,
        batch_size: int = 32,
        shuffle: bool = True,
        num_workers: int = 0,
        pin_memory: bool = True,
    ):
        r"""
        Train the autoencoder model. The logs are stored in the directory specified by BASEDIR with tensorboard format.

        Args:
            train_dataset (torch.utils.data.Dataset): Training dataset.
            eval_dataset (torch.utils.data.Dataset): Evaluation dataset.
            epochs (int): Number of epochs to train the model. Default is ``100``.
            callback: Callback object. Default is ``None``.
            lr (float): Learning rate. Default is ``1e-3``.
            BASEDIR (str): Directory to save the model. Default is ``"./"``.
            reduction (str): Reduction method for the loss function. Default is ``"mean"``.
            lr_decay (float): Learning rate decay. Default is ``0.999``.
            batch_size (int): Batch size. Default is ``32``.
            shuffle (bool): Whether to shuffle the dataset or not. Default is ``True``.
            num_workers (int): Number of workers for the Dataloader. Default is ``0``.
            pin_memory (bool): Pin memory for Dataloader. Default is ``True``.
        """
        dataloader_params = {
            "batch_size": batch_size,
            "shuffle": shuffle,
            "num_workers": num_workers,
            "pin_memory": pin_memory,
        }
        train_data = DataLoader(train_dataset, **dataloader_params)
        eval_data  = DataLoader(eval_dataset, **dataloader_params)
        # Initialization
        prev_train_loss = 1e99
        writer = SummaryWriter(BASEDIR)
        optimizer = torch.optim.AdamW(self.parameters(), lr=lr)
        scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=lr_decay)
        # Training loop
        for epoch in range(epochs):
            self.train()
            num_batches = 0
            tr_loss = 0
            for batch0 in train_data:
                batch = batch0.to(self._device)
                recon, _ = self(batch)
                loss = self._lossfunc(batch, recon, reduction)
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                tr_loss += loss.item()
                num_batches += 1
            tr_loss /= num_batches
            # Validation phase
            if eval_dataset is not None:
                with torch.no_grad():
                    val_batches = 0
                    va_loss = 0
                    for val_batch0 in eval_data:
                        val_batch = val_batch0.to(self._device)
                        val_recon, _ = self(val_batch)
                        vali_loss = self._lossfunc(val_batch, val_recon, reduction)
                        va_loss += vali_loss.item()
                        val_batches += 1
                    va_loss /= val_batches
            # Logging
            writer.add_scalar("Loss/train", tr_loss, epoch + 1)
            writer.add_scalar("Loss/vali", va_loss, epoch + 1)
            # Early stopping
            if callback and callback.early_stop(va_loss, prev_train_loss, tr_loss):
                print(f'Early Stopper Activated at epoch {epoch}', flush=True)
                break
            prev_train_loss = tr_loss
            print(f'Epoch [{epoch+1} / {epochs}] average training loss: {tr_loss:.5e} | average validation loss: {va_loss:.5e}', flush=True)            
            # Learning rate scheduling
            scheduler.step()

        # Cleanup
        writer.flush()
        writer.close()
        torch.save(self.state_dict(), f'{BASEDIR}/model_state.pth')

    def reconstruct(self, dataset: torch.utils.data.Dataset):
        r"""
        Reconstruct the dataset using the trained autoencoder model. It prints the energy, mean, and fluctuation of the reconstructed dataset.

        Args:
            dataset (torch.utils.data.Dataset): Dataset to reconstruct.

        Returns:
            np.ndarray: Reconstructed dataset.
        """
        ## Compute reconstruction and its accuracy
        num_samples = len(dataset)
        ek = np.zeros(num_samples)
        mu = np.zeros(num_samples)
        si = np.zeros(num_samples)
        rec = torch.zeros((self.inp_chan, self.N, num_samples), device=self._device)

        loader = torch.utils.data.DataLoader(dataset, batch_size=num_samples, shuffle=False)

        with torch.no_grad():
            ## Energy recovered in reconstruction
            for energy_batch in loader:
                energy_batch = energy_batch.to(self._device)
                x_recon,_ = self(energy_batch)

                for i in range(num_samples):
                    x_recchan = x_recon[i]
                    rec[:, :, i] = x_recchan.view(self.inp_chan, self.N)

                    x = energy_batch[i].view(self.inp_chan * self.N)
                    xr = rec[:, :, i].view(self.inp_chan * self.N)

                    ek[i] = torch.sum((x - xr) ** 2) / torch.sum(x ** 2)
                    mu[i] = 2 * torch.mean(x) * torch.mean(xr) / (torch.mean(x) ** 2 + torch.mean(xr) ** 2)
                    si[i] = 2 * torch.std(x) * torch.std(xr) / (torch.std(x) ** 2 + torch.std(xr) ** 2)

        energy = (1 - np.mean(ek)) * 100
        print('Recovered energy %.2f' % energy)
        print('Recovered mean %.2f' % (np.mean(mu) * 100))
        print('Recovered fluct %.2f' % (np.mean(si) * 100))

        return rec.cpu().numpy()
    
    def latent_space(self, dataset: torch.utils.data.Dataset):
        r"""
        Compute the latent space of the elements of a given dataset.

        Args:
            dataset (torch.utils.data.Dataset): Dataset to compute the latent space.

        Returns:
            np.ndarray: Latent space of the dataset elements.
        """
        # Compute latent vectors
        loader = torch.utils.data.DataLoader(dataset, batch_size=len(dataset), shuffle=False)
        with torch.no_grad():
            instant  = iter(loader)
            batch    = next(instant)
            batch    = batch.to(self._device)
            _,z = self(batch)
        return z

    def decode(self, z):
        r"""
        Decode the latent space to the original space.

        Args:
            z (np.ndarray): Element of the latent space.

        Returns:
            np.ndarray: Decoded latent space.
        """
        zt  = torch.tensor(z, dtype=torch.float32)
        var = self.decoder(zt)
        var = var.cpu()
        varr = np.zeros((self.N,var.shape[0]),dtype=float)
        for it in range(var.shape[0]):
            varaux = var[it,0,:,:].detach().numpy()
            varr[:,it] = varaux.reshape((self.N,), order='C')
        return varr 

## Wrapper of a variational autoencoder
class VariationalAutoencoder(Autoencoder):
    r"""
    Variational Autoencoder class for neural network module. The model is based on the PyTorch.

    Args:
        latent_dim (int): Dimension of the latent space.
        in_shape (tuple): Shape of the input data.
        input_channels (int): Number of input channels.
        encoder (torch.nn.Module): Encoder model.
        decoder (torch.nn.Module): Decoder model.
        device (str): Device to run the model. Default is 'cuda' if available, otherwise 'cpu'.

    """
    def __init__(self, latent_dim, in_shape, input_channels, encoder, decoder, device=DEVICE):
        super(VariationalAutoencoder, self).__init__(latent_dim, in_shape, input_channels, encoder, decoder, device)

    def _reparamatrizate(self, mu, logvar):
        std = torch.exp(0.5*logvar)
        epsilon = torch.randn_like(std)  #we create a normal distribution (0 ,1 ) with the dimensions of std        
        sample = mu + std*epsilon
        return  sample
             
    def _kld(self, mu, logvar):
        mum     = torch.mean(mu, axis=0)
        logvarm = torch.mean(logvar, axis=0)
        return 0.5*torch.sum(1 + logvar - mum**2 - logvarm.exp())
    
    def forward(self, x):
        mu, logvar = self.encoder(x)
        z = self._reparamatrizate(mu, logvar)
        recon = self.decoder(z)
        return recon, mu, logvar, z
    
    @cr('VAE.fit')
    def fit(
        self,
        train_dataset,
        eval_dataset=None,
        betasch=None,
        epochs=1000,
        callback=None,
        lr=1e-4,
        BASEDIR="./",
        batch_size=32,
        shuffle=True,
        num_workers=0,
        pin_memory=True,
    ):
        r"""
        Train the variational autoencoder model. The logs are stored in the directory specified by BASEDIR with tensorboard format.

        Args:
            train_dataset (torch.utils.data.Dataset): Training dataset.
            eval_dataset (torch.utils.data.Dataset): Evaluation dataset.
            epochs (int): Number of epochs to train the model. Default is ``100``.
            callback: Callback object to change the value of beta during training. Default is ``None``.
            lr (float): Learning rate. Default is ``1e-3``.
            BASEDIR (str): Directory to save the model. Default is ``"./"``.
            reduction (str): Reduction method for the loss function. Default is ``"mean"``.
            lr_decay (float): Learning rate decay. Default is ``0.999``.
            batch_size (int): Batch size. Default is ``32``.
            shuffle (bool): Whether to shuffle the dataset or not. Default is ``True``.
            num_workers (int): Number of workers for the Dataloader. Default is ``0``.
            pin_memory (bool): Pin memory for Dataloader. Default is ``True``.
        """

        dataloader_params = {
            "batch_size": batch_size,
            "shuffle": shuffle,
            "num_workers": num_workers,
            "pin_memory": pin_memory,
        }
        train_data = DataLoader(train_dataset, **dataloader_params)
        eval_data  = DataLoader(eval_dataset, **dataloader_params)
        prev_train_loss = 1e99
        writer    = SummaryWriter(BASEDIR)
        optimizer = torch.optim.Adam(self.parameters(), lr=lr, weight_decay=0, amsgrad=False if self._device == "cpu" else True, fused=False if self._device == "cpu" else True)
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, epochs, eta_min=lr*1e-3)
        scaler    = GradScaler()
        for epoch in range(epochs):
            ## Training
            self.train()
            tr_loss = 0
            mse     = 0
            kld     = 0
            beta    = betasch.getBeta(epoch) if betasch is not None else 0
            for batch0 in train_data:
                batch = batch0.to(self._device)
                optimizer.zero_grad()
                with autocast(device_type=self._device):
                    recon, mu, logvar, _ = self(batch)
                    mse_i = self._lossfunc(batch, recon, reduction='sum')
                    kld_i = self._kld(mu,logvar)
                    loss  = mse_i - beta*kld_i
                scaler.scale(loss).backward()
                scaler.step(optimizer)
                scaler.update()
                tr_loss += loss.item()
                mse     += mse_i.item()
                kld     += kld_i.item()
            num_batches = len(train_data)
            tr_loss /= num_batches
            mse /= num_batches
            kld /= num_batches

            ## Validation
            self.eval()
            va_loss     = 0
            with torch.no_grad():
                for val_batch0 in eval_data:
                    val_batch = val_batch0.to(self._device)
                    with autocast(device_type=self._device):
                        val_recon, val_mu, val_logvar, _ = self(val_batch)
                        mse_i     = self._lossfunc(val_batch, val_recon, reduction='sum')
                        kld_i     = self._kld(val_mu,val_logvar)
                        vali_loss = mse_i - beta*kld_i
                    va_loss  += vali_loss.item()

            num_batches = len(eval_data)
            va_loss    /=num_batches
            writer.add_scalar("Loss/train",tr_loss,epoch+1)
            writer.add_scalar("Loss/vali", va_loss,epoch+1)
            writer.add_scalar("Loss/mse",  mse,    epoch+1)
            writer.add_scalar("Loss/kld",  kld,    epoch+1)

            if callback is not None:
                if callback.early_stop(va_loss, prev_train_loss, tr_loss):
                    pprint(0, 'Early Stopper Activated at epoch %i' %epoch, flush=True)
                    break
            prev_train_loss = tr_loss   
            pprint(0, 'Epoch [%d / %d] average training loss: %.5e (MSE = %.5e KLD = %.5e) | average validation loss: %.5e' % (epoch+1, epochs, tr_loss, mse, kld, va_loss), flush=True)
            # Learning rate scheduling
            scheduler.step()

        writer.flush()
        writer.close()
        torch.save(self.state_dict(), '%s/model_state' % BASEDIR)

    @cr('VAE.reconstruct')
    def reconstruct(self, dataset):
        r"""
        Reconstruct the dataset using the trained variational autoencoder model. It prints the energy, mean, and fluctuation of the reconstructed dataset.

        Args:
            dataset (torch.utils.data.Dataset): Dataset to reconstruct.

        Returns:
            np.ndarray: Reconstructed dataset.
        """
        ## Compute reconstruction and its accuracy
        num_samples = len(dataset)
        ek = np.zeros(num_samples)
        mu = np.zeros(num_samples)
        si = np.zeros(num_samples)
        rec = torch.zeros((self.inp_chan, self.N, num_samples), device=self._device)

        loader = torch.utils.data.DataLoader(dataset, batch_size=num_samples, shuffle=False)

        with torch.no_grad():
            ## Energy recovered in reconstruction
            for energy_batch in loader:
                energy_batch = energy_batch.to(self._device)
                x_recon,_,_,_ = self(energy_batch)

                for i in range(num_samples):
                    x_recchan = x_recon[i]
                    rec[:, :, i] = x_recchan.view(self.inp_chan, self.N)

                    x = energy_batch[i].view(self.inp_chan * self.N)
                    xr = rec[:, :, i].view(self.inp_chan * self.N)

                    ek[i] = torch.sum((x - xr) ** 2) / torch.sum(x ** 2)
                    mu[i] = 2 * torch.mean(x) * torch.mean(xr) / (torch.mean(x) ** 2 + torch.mean(xr) ** 2)
                    si[i] = 2 * torch.std(x) * torch.std(xr) / (torch.std(x) ** 2 + torch.std(xr) ** 2)

        energy = (1 - np.mean(ek)) * 100
        print('Recovered energy %.2f' % energy)
        print('Recovered mean %.2f' % (np.mean(mu) * 100))
        print('Recovered fluct %.2f' % (np.mean(si) * 100))

        return rec.cpu().numpy()
  
    def correlation(self, dataset):
        r"""
        Compute the correlation between the latent variables of the given dataset.

        Args:
            dataset (torch.utils.data.Dataset): Dataset to compute the correlation.

        Returns:
            np.ndarray: Correlation between the latent variables.
        """
        ##  Compute correlation between latent variables
        loader = torch.utils.data.DataLoader(dataset, batch_size=len(dataset), shuffle=False)
        with torch.no_grad():
            instant  = iter(loader)
            batch    = next(instant)
            batch    = batch.to(self._device)
            _,_,_, z = self(batch)
            np.save('z.npy',z.cpu())
            corr = np.corrcoef(z.cpu(),rowvar=False)
        detR = np.linalg.det(corr)*100
        print('Orthogonality between modes %.2f' % (detR))
        return corr, detR#.reshape((self.lat_dim*self.lat_dim,))
    
    def modes(self):
        r"""
        Compute the modes of the latent space.

        Returns:
            np.ndarray: Modes of the latent space.
        """
        zmode = np.diag(np.ones((self.lat_dim,),dtype=float))
        zmodt = torch.tensor(zmode, dtype=torch.float32)
        zmodt = zmodt.to(self._device)
        modes = self.decoder(zmodt)
        mymod = np.zeros((self.N,self.lat_dim),dtype=float)
        modes = modes.cpu()
        for imode in range(self.lat_dim):
            modesr = modes[imode,0,:,:].detach().numpy()
            mymod[:,imode] = modesr.reshape((self.N,), order='C')
        return mymod.reshape((self.N*self.lat_dim,),order='C')

    def latent_space(self, dataset):
        r"""
        Compute the latent space of the elements of a given dataset.

        Args:
            dataset (torch.utils.data.Dataset): Dataset to compute the latent space.

        Returns:
            np.ndarray: Latent space of the dataset elements.
        """
        # Compute latent vectors
        loader = torch.utils.data.DataLoader(dataset, batch_size=len(dataset), shuffle=False)
        with torch.no_grad():
            instant  = iter(loader)
            batch    = next(instant)
            batch    = batch.to(self._device)
            _,_,_, z = self(batch)
        return z

    def fine_tune(self, train_dataset, shape_, eval_dataset=None, epochs=1000, callback=None, lr=1e-4, BASEDIR='./', **dataloader_params):
        train_data = DataLoader(torch.from_numpy(train_dataset).to(torch.float32), **dataloader_params)
        eval_data  = DataLoader(torch.from_numpy(eval_dataset).to(torch.float32), **dataloader_params)
        prev_train_loss = 1e99
        writer    = SummaryWriter(BASEDIR)
        decoder_model = self.decoder
        optimizer = torch.optim.Adam(decoder_model.parameters(), lr=lr, weight_decay=0, amsgrad=False if self._device == "cpu" else True, fused=False if self._device == "cpu" else True)
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, epochs, eta_min=lr*1e-3)
        scaler    = GradScaler()
        
        for epoch in range(epochs):
            ## Training
            decoder_model.train()
            tr_loss = 0
            for batch0 in train_data:
                batch = batch0.to(self._device)
                optimizer.zero_grad()
                with autocast(device_type=self._device):
                    in_data = batch[:, :self.lat_dim]
                    recon   = decoder_model(in_data)
                    loss    = self._lossfunc(torch.reshape(batch[:, self.lat_dim:], recon.shape), recon, reduction='sum')
                    
                scaler.scale(loss).backward()
                scaler.step(optimizer)
                scaler.update()
                tr_loss += loss.item()
            
            num_batches = len(train_data)
            tr_loss /= num_batches
            
            ## Validation
            decoder_model.eval()
            va_loss     = 0
            with torch.no_grad():
                for val_batch0 in eval_data:
                    val_batch = val_batch0.to(self._device)
                    with autocast(device_type=self._device):
                        val_in_data   = val_batch[:, :self.lat_dim]
                        val_recon     = decoder_model(val_in_data)
                        vali_loss     = self._lossfunc(torch.reshape(batch[:, self.lat_dim:], val_recon.shape), val_recon, reduction='sum')
                        
                    va_loss  += vali_loss.item()

            num_batches = len(eval_data)
            va_loss    /=num_batches
            writer.add_scalar("Ft/Loss/train",tr_loss,epoch+1)
            writer.add_scalar("Ft/Loss/vali", va_loss,epoch+1)
            
            if callback is not None:
                if callback.early_stop(va_loss, prev_train_loss, tr_loss):
                    pprint(0, 'Early Stopper Activated at epoch %i' %epoch, flush=True)
                    break
            prev_train_loss = tr_loss   
            pprint(0, 'Epoch [%d / %d] average training loss: %.5e | average validation loss: %.5e' % (epoch+1, epochs, tr_loss, va_loss), flush=True)
            # Learning rate scheduling
            scheduler.step()

        writer.flush()
        writer.close()
        torch.save(decoder_model.state_dict(), '%s/decoder_state' % BASEDIR)
        
        self.decoder.load_state_dict(torch.load('%s/decoder_state' % BASEDIR, weights_only=True))

        return 0

    def decode(self, z):
        r"""
        Decode a latent space element to the original space.

        Args:
            z (np.ndarray): Element of the latent space.

        Returns:
            np.ndarray: Decoded latent space.
        """
        zt  = torch.tensor(z, dtype=torch.float32)
        var = self.decoder(zt)
        var = var.cpu()
        varr = np.zeros((self.N,var.shape[0]),dtype=float)
        for it in range(var.shape[0]):
            varaux = var[it,0,:,:].detach().numpy()
            varr[:,it] = varaux.reshape((self.N,), order='C')
        return varr 


#!/bin/env python
# 
# Test pyLOM multi-gpu using cupy
# Strong scalability test, the global size of the problem is kept constant
#
# Matrix size guidelines depending on the machine tested:
#
#### A node of MareNostrum 5 (4 NVIDIA H100 GPUs per node with 64 Gb of VRAM each) can handle the following sizes:
######## M = 8e6, N = 500
######## M = 4e6, N = 1000
######## M = 2e6, N = 2000
#
#### A single GPU of Juno 3 (48 Gb of VRAM) can handle the following sizes:
######## M = 8e6, N = 250
######## M = 4e6, N = 500
######## M = 2e6, N = 1000
#
#### Add any other tested machine ####
#
# Last revision: 28/02/2025
from __future__ import print_function, division

import numpy as np, cupy as cp
import pyLOM
import sys

from pyLOM.utils.mpi import MPI_RANK, MPI_SIZE

# Run on GPU. Adjust accordingly to the number of GPUs per node of your machine
pyLOM.gpu_device(gpu_per_node=4) 

# Define the global matrix size according to the capabilities of your machine
M = int(np.double(sys.argv[1]))
N = int(sys.argv[2])

# Split the number of rows between the number of processors
istart, iend = pyLOM.utils.worksplit(0,M,MPI_RANK,nWorkers=MPI_SIZE)
Mi = iend-istart

# Generate random matrix
Ai = cp.random.rand(Mi,N).astype(cp.float32)

## Add mpi_barrier for timing consistency
pyLOM.utils.mpi_barrier()

# Run pyLOM TSQR-SVD algorithm
Ui, S, V = pyLOM.math.tsqr_svd(Ai)

# Print results
pyLOM.pprint(0,f'RUNNING WITH {MPI_SIZE} GPUs')
pyLOM.cr_info()



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# DMD Module
#
# Last rev: 30/09/2021

# Functions coming from DMD
from .wrapper import run, frequency_damping, reconstruction_jovanovic
from .utils   import extract_modes, save, load
from .plots   import plotMode, ritzSpectrum, amplitudeFrequency, dampingFrequency, plotResidual, plotSnapshot

del wrapper



#!/usr/bin/env python
#
# Example of 2D-VAE.
#
# Last revision: 24/09/2024
from __future__ import print_function, division

import mpi4py
mpi4py.rc.recv_mprobe = False

import os, numpy as np
import pyLOM, pyLOM.NN


## Set device
device = pyLOM.NN.select_device("cpu") # Force CPU for this example, if left in blank it will automatically select the device


## Specify autoencoder parameters
nlayers     = 5
channels    = 64
lat_dim     = 5
beta        = 0
beta_start  = 0
beta_wmup   = 0
kernel_size = 4
nlinear     = 512
padding     = 1
activations = [pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu(), pyLOM.NN.relu()]


## Load pyLOM dataset and set up results output
BASEDIR = './DATA/'
CASESTR = 'CYLINDER'
DSETDIR = os.path.join(BASEDIR,f'{CASESTR}.h5')
RESUDIR = 'vae_beta_%.2e_ld_%i' % (beta, lat_dim)
pyLOM.NN.create_results_folder(RESUDIR)


## Mesh size (HARDCODED BUT MUST BE INCLUDED IN PYLOM DATASET)
n0h = 449
n0w = 199
nh  = 448
nw  = 192


## Create a torch dataset
m    = pyLOM.Mesh.load(DSETDIR)
d    = pyLOM.Dataset.load(DSETDIR,ptable=m.partition_table)
u_x  = d['VELOX']
u_m  = pyLOM.math.temporal_mean(u_x)
u_xm = pyLOM.math.subtract_mean(u_x, u_m)
time = d.get_variable('time')
td   = pyLOM.NN.Dataset((u_xm,), (n0h, n0w))
td.crop(nh, nw)


## Set and train the variational autoencoder
betasch    = pyLOM.NN.betaLinearScheduler(0., beta, beta_start, beta_wmup)
encoder    = pyLOM.NN.Encoder2D(nlayers, lat_dim, nh, nw, td.num_channels, channels, kernel_size, padding, activations, nlinear, vae=True)
decoder    = pyLOM.NN.Decoder2D(nlayers, lat_dim, nh, nw, td.num_channels, channels, kernel_size, padding, activations, nlinear)
model      = pyLOM.NN.VariationalAutoencoder(lat_dim, (nh, nw), td.num_channels, encoder, decoder, device=device)
early_stop = pyLOM.NN.EarlyStopper(patience=5, min_delta=0.02)

pipeline = pyLOM.NN.Pipeline(
    train_dataset = td,
    test_dataset  = td,
    model=model,
    training_params={
        "batch_size": 4,
        "epochs": 100,
        "lr": 1e-4,
        "betasch": betasch,
        "BASEDIR": RESUDIR
    },
)
pipeline.run()


## Reconstruct dataset and compute accuracy
rec = model.reconstruct(td)
rd  = pyLOM.NN.Dataset((rec,), (nh, nw))
rd.pad(n0h, n0w)
td.pad(n0h, n0w)
d.add_field('urec',1,rd[:,0,:,:].numpy().reshape((len(time),n0w*n0h)).T)
d.add_field('utra',1,td[:,0,:,:].numpy().reshape((len(time),n0w*n0h)).T)
pyLOM.io.pv_writer(m,d,'reco',basedir=RESUDIR,instants=np.arange(time.shape[0],dtype=np.int32),times=time,vars=['urec', 'VELOX', 'utra'],fmt='vtkh5')
pyLOM.NN.plotSnapshot(m,d,vars=['urec'],instant=0,component=0,cmap='jet',cpos='xy')
pyLOM.NN.plotSnapshot(m,d,vars=['utra'],instant=0,component=0,cmap='jet',cpos='xy')


pyLOM.cr_info()


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# DMD general utilities.
#
# Last rev: 27/01/2023
from __future__ import print_function, division

import numpy as np

from ..utils.gpu import cp
from ..          import inp_out as io
from ..utils     import cr_nvtx as cr, gpu_to_cpu


@cr('DMD.extract_modes')
def extract_modes(Phi,ivar,npoints,real=True,modes=[],reshape=True):
	'''
	Extract modes for a certain variables
	'''
	p = cp if type(Phi) is cp.ndarray else np
	nvars = Phi.shape[0]//npoints
	# Define modes to extract
	if len(modes) == 0: modes = p.arange(1,Phi.shape[1]+1,dtype=p.int32)
	# Allocate output array
	out =p.zeros((npoints,len(modes)),p.double if Phi.dtype == p.complex128 else p.float32)
	for i,m in enumerate(modes):
		out[:,i] = Phi[ivar-1:nvars*npoints:nvars,m-1].real if real else Phi[ivar-1:nvars*npoints:nvars,m-1].imag
	# Return reshaped output
	return out.reshape((len(modes)*npoints,),order='C') if reshape else out


@cr('DMD.save',color='blue')
def save(fname,muReal,muImag,Phi,bJov,ptable,nvars=1,pointData=True,mode='w'):
	'''
	Store DMD variables in serial or parallel
	according to the partition used to compute the DMD.
	'''
	io.h5_save_DMD(fname,gpu_to_cpu(muReal),gpu_to_cpu(muImag),gpu_to_cpu(Phi),gpu_to_cpu(bJov),ptable,nvars=nvars,pointData=pointData,mode=mode)


@cr('DMD.load',color='blue')
def load(fname,vars=['Phi','mu','bJov','delta','omega'],nmod=-1,ptable=None):
	'''
	Load DMD variables in serial or parallel
	according to the partition used to compute the DMD.
	'''
	return io.h5_load_DMD(fname,vars,nmod,ptable)



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# GPR Module
#
# Last rev: 19/02/2025

import numpy as np, GPy

# Multi Fidelity GPR Model
from emukit.multi_fidelity.convert_lists_to_array import convert_x_list_to_array, convert_xy_lists_to_arrays
from emukit.multi_fidelity.kernels                import LinearMultiFidelityKernel
from emukit.multi_fidelity.models                 import GPyLinearMultiFidelityModel
from emukit.model_wrappers.gpy_model_wrappers     import GPyMultiOutputWrapper

from ..utils  import cr_nvtx as cr, raiseError, pprint


# KernelSelector: instantiates kernels with bounds and provides information
class KernelSelector:
    """
    A kernel selector with a limited list of available kernels.
    Allows retrieving the parameters of each kernel without needing to instantiate it.
    """

    def __init__(self, input_dim: int):
        self.input_dim = input_dim
        self.available_kernels = {
            "RBF": GPy.kern.RBF,
            "Matern32": GPy.kern.Matern32,
            "Matern52": GPy.kern.Matern52,
            "Exponential": GPy.kern.Exponential,
            "Linear": GPy.kern.Linear,
            "RatQuad": GPy.kern.RatQuad,
            "Poly": GPy.kern.Poly,
            "PeriodicExponential": GPy.kern.PeriodicExponential,
            "StdPeriodic": GPy.kern.StdPeriodic,
            "White": GPy.kern.White,
        }
        # Dynamically add functions to create each kernel.
        for name in self.available_kernels:
            setattr(self, name, self._make_kernel_func(name))

    def _make_kernel_func(self, kernel_name: str):
        """
        Creates a function that generates a kernel with the specified name.
        Allows passing parameters and bounds.
        """

        def kernel_func(**kwargs):
            limits = kwargs.pop("limits", None)
            kernel_class = self.available_kernels[kernel_name]
            kernel = kernel_class(self.input_dim, **kwargs)
            if limits:
                for param, (lower, upper) in limits.items():
                    if hasattr(kernel, param):
                        kernel[param].constrain_bounded(lower, upper)
                    else:
                        raiseError(
                            f"The kernel '{kernel_name}' does not have the parameter '{param}'."
                        )
            return kernel

        return kernel_func

    def get_kernel_parameters(self, kernel_name: str):
        """
        Returns the parameter names of the kernel and indicates if it supports ARD.
        """
        kernel_class = self.available_kernels.get(kernel_name)
        if kernel_class is None:
            raiseError(f"Kernel class '{kernel_name}' not found.")
        try:
            kernel_instance = kernel_class(self.input_dim)
        except TypeError:
            raiseError(
                f"Could not instantiate '{kernel_name}' with input_dim={self.input_dim}."
            )
        param_names = list(kernel_instance.parameter_names())
        has_ard = hasattr(kernel_instance, "ARD") or (
            "lengthscale" in param_names and kernel_instance.lengthscale.size > 1
        )
        return {"parameters": param_names, "ARD": has_ard}

    def __dir__(self):
        """
        Allows dir(obj) to list the available kernels.
        """
        return list(self.available_kernels.keys()) + super().__dir__()


# Base class with common utilities
class GPRBase:
    def __init__(self):
        self.model = None
        self.input_dim = None

    @staticmethod
    def ensure_column_matrix(arr):
        """
        Ensures that the array is in column matrix shape.
        """
        arr = np.asarray(arr)
        if arr.ndim == 1:
            arr = arr[:, np.newaxis]
        if arr.shape[0] < arr.shape[1]:
            arr = arr.T
        return arr


# Single Fidelity GPR Model
class SF_GPR(GPRBase):
    """
    Class for Single Fidelity Gaussian Process Regression (GPR) using GPy.

    The interface is similar to sklearn:
      - fit(X_train, y_train, kernel, noise_var, num_restarts, verbose)
      - predict(X_test) returns a dictionary with 'mean' and 'std'

    It ensures that the data are in column matrix shape.
    """

    def __init__(self, input_dim=None):
        """
        If input_dim is provided, a KernelSelector is created immediately.
        """
        super().__init__()
        self._kernel_selector = None
        self._kernel = None
        if input_dim is not None:
            self.input_dim = input_dim
            self._kernel_selector = KernelSelector(input_dim)

    @property
    def kernel(self):
        """
        Allows access to the KernelSelector to create kernels (e.g., model.kernel.RBF(...)).
        """
        if self._kernel_selector is None:
            raiseError(
                "Provide input_dim in the constructor before accessing the kernel."
            )
        return self._kernel_selector

    @cr('SF_GPR.fit')
    def fit(
        self, X_train, y_train, kernel, noise_var=None, num_restarts=5, verbose=True
    ):
        """
        Fits the GPR model:
          - X_train, y_train: training data (converted to column matrix).
          - kernel: expects a kernel already created.
          - noise_var: if provided (not None), the Gaussian noise variance is fixed;
                       otherwise, it is left free for optimization.
          - num_restarts and verbose: parameters for optimization.
        """
        self.train_features = self.ensure_column_matrix(X_train)
        self.train_labels = self.ensure_column_matrix(y_train)
        # If input_dim was not provided in the constructor, infer it from the data
        if self.input_dim is None:
            self.input_dim = self.train_features.shape[1]
        if self._kernel_selector is None:
            self._kernel_selector = KernelSelector(self.input_dim)
        self._kernel = kernel
        self.model = GPy.models.GPRegression(
            self.train_features, self.train_labels, self._kernel
        )
        if noise_var is not None:
            self.model.Gaussian_noise.fix(noise_var)
        self.model.optimize_restarts(num_restarts, verbose=verbose)
        return self

    def predict(self, X_test):
        """
        Makes predictions with the model using the test data X_test.
        Ensures that X_test is in column matrix shape and returns a dictionary with 'mean' and 'std'.
        """
        if self.model is None:
            raiseError("Fit the model before predicting.")
        X_test = self.ensure_column_matrix(X_test)
        mean, var = self.model.predict(X_test)
        return {"mean": mean, "std": np.sqrt(var)}

    def display_model(self):
        """
        Displays the model summary.
        """
        pprint(0,self.model)


class MF_GPR(GPRBase):
    """
    Model for multi-fidelity Gaussian Process Regression.

    Interface similar to sklearn:
      - fit(train_features_list, train_labels_list, kernels, noise_vars, num_restarts, verbose)
      - predict(predict_features_list) returns a dictionary with predictions for each fidelity.

    It ensures that each array is in column matrix shape.
    """

    def __init__(self, input_dim=None):
        super().__init__()
        self.train_features_list = None
        self.train_labels_list = None
        self._kernel_selector = None
        if input_dim is not None:
            self.input_dim = input_dim
            self._kernel_selector = KernelSelector(input_dim)
        self.n_fidelities = None
        self.kernel_MF = None
        self.wrapper = None

    @property
    def kernel(self):
        """
        Allows access to the KernelSelector to create kernels for each fidelity.
        """
        if self._kernel_selector is None:
            raiseError(
                "Provide input_dim in the constructor before accessing the kernel."
            )
        return self._kernel_selector

    @cr('MF_GPR.fit')
    def fit(
        self,
        train_features_list,
        train_labels_list,
        kernels,
        noise_vars=None,
        num_restarts=5,
        verbose=True,
    ):
        """
        Fits the multi-fidelity model:
          - train_features_list, train_labels_list: lists of training arrays (one per fidelity)
          - kernels: list of kernels (one per fidelity), which are combined internally
          - noise_vars: float or list of floats to fix the noise; if None, noise is free for optimization.
          - num_restarts and verbose: parameters for optimization.
        """
        self.train_features_list = [
            self.ensure_column_matrix(x) for x in train_features_list
        ]
        self.train_labels_list = [
            self.ensure_column_matrix(y) for y in train_labels_list
        ]
        self.input_dim = self.train_features_list[0].shape[1]
        self.n_fidelities = len(self.train_features_list)
        self.X_MF_upper, self.Y_MF_upper = convert_xy_lists_to_arrays(
            self.train_features_list, self.train_labels_list
        )
        if self._kernel_selector is None:
            self._kernel_selector = KernelSelector(self.input_dim)
        if not isinstance(kernels, list) or len(kernels) != self.n_fidelities:
            raiseError(
                "Provide a list of kernels equal in length to the number of fidelities."
            )
        self.kernel_MF = LinearMultiFidelityKernel(kernels)
        self.model = GPyLinearMultiFidelityModel(
            self.X_MF_upper,
            self.Y_MF_upper,
            self.kernel_MF,
            n_fidelities=self.n_fidelities,
        )
        self.wrapper = GPyMultiOutputWrapper(
            self.model,
            self.n_fidelities,
            n_optimization_restarts=num_restarts,
            verbose_optimization=verbose,
        )
        if noise_vars is not None:
            self._fix_noise(noise_vars)
        self.wrapper.optimize()
        return self

    def _fix_noise(self, noise_vars):
        """
        Internal method to fix the Gaussian noise for each fidelity.
        noise_vars: float or list of floats.
        """
        noise_params = [
            p
            for p in self.wrapper.gpy_model.mixed_noise.parameters
            if "Gaussian_noise" in p.name
        ]
        if isinstance(noise_vars, (int, float)):
            noise_vars = [noise_vars] * len(noise_params)
        if len(noise_vars) != len(noise_params):
            raiseError(
                f"Expected {len(noise_params)} noise variances, but got {len(noise_vars)}."
            )
        for param, noise_var in zip(noise_params, noise_vars):
            param.fix(noise_var)

    def predict(self, predict_features_list):
        """
        Performs prediction for each fidelity using the provided test data.
        Expects a list of arrays (one per fidelity); ensures that each array is in column matrix shape.
        Returns a dictionary with predictions for each fidelity (keys 'fidelity_1', etc.).
        """
        if not isinstance(predict_features_list, list):
            raiseError(
                "predict_features_list must be a list of arrays (one per fidelity)."
            )
        predict_features_list = [
            self.ensure_column_matrix(x) for x in predict_features_list
        ]
        x_mf_predict = convert_x_list_to_array(predict_features_list)
        results = {}
        start = 0
        for i, x in enumerate(predict_features_list):
            n_points = x.shape[0]
            Xi = x_mf_predict[start : start + n_points]
            mean, var = self.wrapper.predict(Xi)
            results[f"fidelity_{i+1}"] = {"mean": mean, "std": np.sqrt(var)}
            start += n_points
        return results

    def display_model(self):
        """
        Displays the multi-fidelity model summary.
        """
        pprint(0,self.wrapper.gpy_model)




#!/usr/bin/env python
#
# Example of SPOD.
#
# Last revision: 19/07/2021
from __future__ import print_function, division

import mpi4py
mpi4py.rc.recv_mprobe = False

import pyLOM

pyLOM.gpu_device(gpu_per_node=4)


## Parameters
DATAFILE = './DATA/Tensor_re280.h5'
VARIABLE = 'VELOC'


## Data loading
m     = pyLOM.Mesh.load(DATAFILE)
d     = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table).to_gpu([VARIABLE])
X     = d[VARIABLE]
t     = d.get_variable('time')
npwin = 40 #Number of snapshots in each window
nolap = 15 #Number of overlapping snapshots between windows


## Run SPOD
L, P, f = pyLOM.SPOD.run(X,t,nDFT=npwin,nolap=nolap,remove_mean=True)
pyLOM.SPOD.save('results.h5',L,P,f,d.partition_table,nvars=3,pointData=d.point)
if pyLOM.utils.is_rank_or_serial(root=0): pyLOM.SPOD.plotSpectra(f, L)


## Dump to ParaView
# Spatial modes
d.add_field('spatial_modes_U',6,pyLOM.SPOD.extract_modes(L,P,1,len(d),modes=[1,2,3,4,5,6]))
pyLOM.io.pv_writer(m,d.to_cpu(['spatial_modes_U']),'modes',basedir='out/modes',instants=[0],times=[0.],vars=['spatial_modes_U'],fmt='vtkh5')
pyLOM.SPOD.plotMode(L,P,f,m,d,1,pointData=d.point,modes=[1,2,3,4,5,6],cpos='xy')


## Show and print timings
pyLOM.cr_info()
pyLOM.show_plots()


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# DMD plotting utilities.
#
# Last rev: 27/10/2021
from __future__ import print_function, division

import numpy as np
import matplotlib.pyplot as plt

from .utils        import extract_modes
from ..vmmath      import fft
from ..utils       import gpu_to_cpu
from ..utils.plots import plotResidual, plotFieldStruct2D, plotSnapshot, plotLayout


def plotMode(L, P, freqs, mesh, dset, ivar, pointData=True, modes=np.array([1],np.int32),**kwargs):
	'''
	Plot the real and imaginary parts of a mode
	'''
	L, P = gpu_to_cpu(L), gpu_to_cpu(P)
	# Extract the modes to be plotted
	npoints = mesh.size(pointData)
	P_modes = extract_modes(L,P,ivar,npoints,modes=modes)
	# Add to the dataset
	dset.add_field('P_MODES',len(modes),P_modes)
	# Loop over the modes
	screenshot = kwargs.pop('screenshot',None)
	off_screen = kwargs.pop('off_screen',False)
	for imode, mode in enumerate(modes):
		if screenshot is not None: kwargs['screenshot'] = screenshot % imode
		plotLayout(mesh,dset,1,1,mode-1,vars=['P_MODES'],title='Mode %d St = %.3f' % (mode-1, np.abs(freqs[mode-1])),off_screen=off_screen,**kwargs)
	# Remove from dataset
	dset.delete('P_MODES')

def plotSpectra(f, L, fig=None, ax=None):
	L, f = gpu_to_cpu(L), gpu_to_cpu(f)
	# Get or recover axis and figure
	if fig is None:
		fig = plt.figure(figsize=(8,6),dpi=100)
	if ax is None:
		ax = fig.add_subplot(1,1,1)
	# Plot
	for ii in range(L.shape[1]):
		ax.loglog(np.sort(f), L[np.argsort(f),ii], 'o-')
	ax.set_xlabel('St')
	ax.set_ylabel(r'$\lambda_i$')
	return fig, ax


#!/usr/bin/env python
#
# Example of 2D-AE.
#
# Last revision: 09/10/2024

import numpy as np
import pyLOM, pyLOM.NN

import matplotlib.pyplot as plt


## Set device
device = pyLOM.NN.select_device("cpu") # Force CPU for this example, if left in blank it will automatically select the device


## Specify autoencoder parameters
nlayers     = 5
channels    = 64
lat_dim     = 5
kernel_size = 4
nlinear     = 512
padding     = 1
activations = [pyLOM.NN.tanh(), pyLOM.NN.tanh(), pyLOM.NN.tanh(), pyLOM.NN.tanh(), pyLOM.NN.tanh(), pyLOM.NN.tanh(), pyLOM.NN.tanh()]


## Load pyLOM dataset and set up results output
BASEDIR  = './DATA'
CASESTR = 'CYLINDER'
DSETDIR = '%s/%s.h5' % (BASEDIR, CASESTR)
RESUDIR = 'ae_ld_%i' % (lat_dim)
pyLOM.NN.create_results_folder(RESUDIR)


## Mesh size
n0h = 449
n0w = 199
nh  = 448
nw  = 192


## Create a torch dataset
m    = pyLOM.Mesh.load(DSETDIR)
d    = pyLOM.Dataset.load(DSETDIR,ptable=m.partition_table)
u    = d['VELOX']
um   = pyLOM.math.temporal_mean(u)
u_x  = pyLOM.math.subtract_mean(u, um)[:,0]
td   = pyLOM.NN.Dataset((u_x,), (n0h, n0w))
td.crop(nh, nw)


## Set and train the variational autoencoder
encoder    = pyLOM.NN.Encoder2D(nlayers, lat_dim, nh, nw, td.num_channels, channels, kernel_size, padding, activations, nlinear)
decoder    = pyLOM.NN.Decoder2D(nlayers, lat_dim, nh, nw, td.num_channels, channels, kernel_size, padding, activations, nlinear)
model      = pyLOM.NN.Autoencoder(lat_dim, (nh, nw), td.num_channels, encoder, decoder, device=device)
early_stop = pyLOM.NN.EarlyStopper(patience=5, min_delta=0.02)

pipeline = pyLOM.NN.Pipeline(
   train_dataset   = td,
   test_dataset    = td,
   model           = model,
   training_params = {
       "batch_size": 16,
       "epochs": 100,
       "lr": 1e-4,
       "callback":early_stop,
       'BASEDIR':RESUDIR,
   },
   
)
pipeline.run()

   
## Reconstruct dataset and compute accuracy
rec = model.reconstruct(td)
rd  = pyLOM.NN.Dataset((rec,), (nh, nw))
rd.pad(n0h,n0w)
td.pad(n0h,n0w)
d.add_field('urec',1,rd[0,0,:,:].numpy().reshape((n0w*n0h,)))
d.add_field('utra',1,td[0,0,:,:].numpy().reshape((n0w*n0h,)))
pyLOM.io.pv_writer(m,d,'reco',basedir=RESUDIR,instants=[0],times=[0.],vars=['urec', 'VELOX', 'utra'],fmt='vtkh5')
pyLOM.NN.plotSnapshot(m,d,vars=['urec'],instant=0,component=0,cmap='jet',cpos='xy')
pyLOM.NN.plotSnapshot(m,d,vars=['utra'],instant=0,component=0,cmap='jet',cpos='xy')

pyLOM.cr_info()


#!/usr/bin/env python
#
# PYLOM Testsuite
# Build jet dataset for the testsuite
#
# Last revision: 03/08/2021
from __future__ import print_function, division

import numpy as np, h5py
import pyLOM


## Parameters
MATFILE = '../DATA/jetLES.mat'
OUTFILE = './DATA/JET.h5'


## Load MAT file
f    = h5py.File(MATFILE,'r')
mat  = { k:np.array(v) for k,v in f.items() }
f.close()
DT   = float(mat['dt'])
time = DT*np.arange(mat['p'].shape[2])+0.04


## Build mesh information dictionary
nx, ny = int(mat['nx']),int(mat['nr']) 
x = np.unique(mat['x'])
y = np.unique(mat['r'])
mesh = pyLOM.Mesh.new_struct2D(nx,ny,x,y,None,None) 
print(mesh)


## Create partition table
ptable = pyLOM.PartitionTable.new(1,mesh.ncells,mesh.npoints)
mesh.partition_table = ptable
print(ptable)


## Obtain pressure array
PRESS = np.ascontiguousarray(mat['p'].reshape((mesh.npoints, 5000)).astype(np.double))


## Create dataset for pyLOM
nsample = 20
d = pyLOM.Dataset(xyz=mesh.xyz, ptable=ptable, order=mesh.pointOrder, point=True,
	# Add the time as the only variable
	vars  = {'time':{'idim':0,'value':time[::nsample]}},
	# Now add all the arrays to be stored in the dataset
	# It is important to convert them as C contiguous arrays
	PRESS = {'ndim':1,'value':PRESS[:,::nsample]},
)
print(d)

mesh.save(OUTFILE) # Store the mesh
d.save(OUTFILE)    # Store dataset


pyLOM.cr_info()


#!/usr/bin/env python
#
# Example of POD.
#
# Last revision: 17/02/2025
from __future__ import print_function, division

import mpi4py
mpi4py.rc.recv_mprobe = False

import numpy as np
import pyLOM

pyLOM.gpu_device(gpu_per_node=4) # Detect GPU configuration


## Parameters
DATAFILE = './DATA/CYLINDER.h5'
VARIABLE = 'VELOC'


## Data loading
m = pyLOM.Mesh.load(DATAFILE)
d = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table).to_gpu([VARIABLE]) # Send to GPU if available
X = d[VARIABLE]
t = d.get_variable('time')


## Run POD
PSI,S,V = pyLOM.POD.run(X,remove_mean=False) # PSI are POD modes
if pyLOM.utils.is_rank_or_serial(root=0): pyLOM.POD.plotResidual(S)
# Truncate to capture a percentage of the energy (r must be negative)
PSI,S,V = pyLOM.POD.truncate(PSI,S,V,r=-0.99)
pyLOM.POD.save('results.h5',PSI,S,V,d.partition_table,nvars=2,pointData=d.point)
# Reconstruct the flow
X_POD = pyLOM.POD.reconstruct(PSI,S,V)
# Compute RMSE
rmse = pyLOM.math.RMSE(X_POD,X)
pyLOM.pprint(0,'RMSE = %e'%rmse)


## Dump to ParaView
# Spatial modes
d.add_field('spatial_modes_U',3,pyLOM.POD.extract_modes(PSI,1,len(d),modes=[1,2,3]))
d.add_field('spatial_modes_V',3,pyLOM.POD.extract_modes(PSI,2,len(d),modes=[1,2,3]))
pyLOM.io.pv_writer(m,d.to_cpu(['spatial_modes_U','spatial_modes_V']),'modes',basedir='out/modes',instants=[0],times=[0.],vars=['spatial_modes_U','spatial_modes_V'],fmt='vtkh5')
pyLOM.POD.plotSnapshot(m,d.to_cpu(['spatial_modes_U']),vars=['spatial_modes_U'],instant=0,component=0,cmap='jet',cpos='xy')

# Temporal evolution
d.add_field('VELOR',2,X_POD)
pyLOM.io.pv_writer(m,d.to_cpu(['VELOC','VELOR']),'flow',basedir='out/flow',instants=np.arange(t.shape[0],dtype=np.int32),times=t,vars=['VELOC','VELOR'],fmt='vtkh5')
pyLOM.POD.plotSnapshot(m,d.to_cpu(['VELOR']),vars=['VELOR'],instant=0,component=0,cmap='jet',cpos='xy')


## Plot POD mode
if pyLOM.utils.is_rank_or_serial(0):
	# 0 - module, 1,2 - components
	pyLOM.POD.plotMode(V[:,:-1],t[:-1],modes=[1,2,3])


## Show and print timings
pyLOM.cr_info()
pyLOM.show_plots()


#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - vector/matrix math.
#
# Last rev: 27/10/2021

cimport cython
cimport numpy as np

import numpy as np

#from libc.complex  cimport creal, cimag
cdef extern from "<complex.h>" nogil:
	float  complex I
	# Decomposing complex values
	float cimagf(float complex z)
	float crealf(float complex z)
	double cimag(double complex z)
	double creal(double complex z)
cdef double complex J = 1j
from libc.stdlib   cimport malloc, free
from libc.string   cimport memcpy, memset
from libc.math     cimport sqrt, atan2
from .cfuncs       cimport real, real_complex, real_full
from .cfuncs       cimport c_stranspose, c_svector_sum, c_svector_norm, c_svector_mean, c_smatmul, c_smatmulp, c_svecmat, c_sinv, c_ssort
from .cfuncs       cimport c_dtranspose, c_dvector_sum, c_dvector_norm, c_dvector_mean, c_dmatmul, c_dmatmulp, c_dvecmat, c_dinv, c_dsort
from .cfuncs       cimport c_cmatmul, c_cmatmulp, c_cvecmat, c_cinv, c_ceigen, c_ccholesky, c_cvandermonde, c_cvandermonde_time, c_csort
from .cfuncs       cimport c_zmatmul, c_zmatmulp, c_zvecmat, c_zinv, c_zeigen, c_zcholesky, c_zvandermonde, c_zvandermonde_time, c_zsort

from ..utils.cr     import cr
from ..utils.errors import raiseError


## Cython functions
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=2] _stranspose(float[:,:] A):
	'''
	Transposed of matrix A
	'''
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] At = np.zeros((n,m),dtype=np.float32)
	c_stranspose(&A[0,0], &At[0,0], m,n)
	return At

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=2] _dtranspose(double[:,:] A):
	'''
	Transposed of matrix A
	'''
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] At = np.zeros((n,m),dtype=np.double)
	c_dtranspose(&A[0,0], &At[0,0], m,n)
	return At

@cr('math.transpose')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def transpose(real[:,:] A):
	r'''
	Transposed of matrix A

	Args:
		A (np.ndarray): Matrix to be transposed
	
	Results
		np.ndarray: Transposed matrix
	'''
	if real is double:
		return _dtranspose(A)
	else:
		return _stranspose(A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef float _svector_sum(float[:] v, int start=0):
	'''
	Sum of a vector
	'''
	cdef int n = v.shape[0]
	cdef float norm = 0.
	norm = c_svector_sum(&v[0],start,n)
	return norm

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef double _dvector_sum(double[:] v, int start=0):
	'''
	Sum of a vector
	'''
	cdef int n = v.shape[0]
	cdef double norm = 0.
	norm = c_dvector_sum(&v[0],start,n)
	return norm

@cr('math.vector_sum')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def vector_sum(real[:] v, int start=0):
	r'''
	Sum of a vector

	Args:
		v (np.ndarray): a vector
		start (int): position of the vector where to start the sum
	
	Result:
		float: sum of the vector
	'''
	if real is double:
		return _dvector_sum(v,start)
	else:
		return _svector_sum(v,start)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef float _svector_norm(float[:] v, int start=0):
	'''
	L2 norm of a vector
	'''
	cdef int n = v.shape[0]
	cdef float norm = 0.
	norm = c_svector_norm(&v[0],start,n)
	return norm

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef double _dvector_norm(double[:] v, int start=0):
	'''
	L2 norm of a vector
	'''
	cdef int n = v.shape[0]
	cdef double norm = 0.
	norm = c_dvector_norm(&v[0],start,n)
	return norm

@cr('math.vector_norm')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def vector_norm(real[:] v, int start=0):
	r'''
	L2 norm of a vector
	
	Args:
		v (np.ndarray): a vector
		start (int): position of the vector where to start the norm
	
	Result:
		float: norm of the vector
	'''
	if real is double:
		return _dvector_norm(v,start)
	else:
		return _svector_norm(v,start)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef float _svector_mean(float[:] v, int start=0):
	'''
	Sum of a vector
	'''
	cdef int n = v.shape[0]
	cdef float norm = 0.
	norm = c_svector_mean(&v[0],start,n)
	return norm

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef double _dvector_mean(double[:] v, int start=0):
	'''
	Sum of a vector
	'''
	cdef int n = v.shape[0]
	cdef double norm = 0.
	norm = c_dvector_mean(&v[0],start,n)
	return norm

@cr('math.vector_mean')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def vector_mean(real[:] v, int start=0):
	r'''
	Mean of a vector

	Args:
		v (np.ndarray): a vector
		start (int): position of the vector where to start the mean
	
	Result:
		float: mean of the vector
	'''
	if real is double:
		return _dvector_mean(v,start)
	else:
		return _svector_mean(v,start)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=2] _smatmul(float[:,:] A, float[:,:] B):
	'''
	Matrix multiplication C = A x B
	'''
	cdef int m = A.shape[0], k = A.shape[1], n = B.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] C = np.zeros((m,n),dtype=np.float32)
	c_smatmul(&C[0,0],&A[0,0],&B[0,0],m,n,k)
	return C

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=2] _dmatmul(double[:,:] A, double[:,:] B):
	'''
	Matrix multiplication C = A x B
	'''
	cdef int m = A.shape[0], k = A.shape[1], n = B.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] C = np.zeros((m,n),dtype=np.double)
	c_dmatmul(&C[0,0],&A[0,0],&B[0,0],m,n,k)
	return C

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex64_t,ndim=2] _cmatmul(np.complex64_t[:,:] A, np.complex64_t[:,:] B):
	'''
	Matrix multiplication C = A x B
	'''
	cdef int m = A.shape[0], k = A.shape[1], n = B.shape[1]
	cdef np.ndarray[np.complex64_t,ndim=2] C = np.zeros((m,n),dtype=np.complex64)
	c_cmatmul(&C[0,0],&A[0,0],&B[0,0],m,n,k)
	return C

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex128_t,ndim=2] _zmatmul(np.complex128_t[:,:] A, np.complex128_t[:,:] B):
	'''
	Matrix multiplication C = A x B
	'''
	cdef int m = A.shape[0], k = A.shape[1], n = B.shape[1]
	cdef np.ndarray[np.complex128_t,ndim=2] C = np.zeros((m,n),dtype=np.complex128)
	c_zmatmul(&C[0,0],&A[0,0],&B[0,0],m,n,k)
	return C

@cr('math.matmul')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def matmul(real_full[:,:] A, real_full[:,:] B):
	r'''
	Matrix multiplication 
	C = A x B

	Args:
		A (np.ndarray): Matrix A (M,Q)
		B (np.ndarray): Matrix B (Q,N)
	
	Result:
		np.ndarray: Resulting matrix C (M,N)
	'''
	# Select type
	if real_full is np.complex128_t:
		return _zmatmul(A,B)
	elif real_full is np.complex64_t:
		return _cmatmul(A,B)
	elif real_full is double:
		return _dmatmul(A,B)
	else:
		return _smatmul(A,B)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=2] _smatmulp(float[:,:] A, float[:,:] B):
	'''
	Matrix multiplication C = A x B
	'''
	cdef int m = A.shape[0], k = A.shape[1], n = B.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] C = np.zeros((m,n),dtype=np.float32)
	c_smatmulp(&C[0,0],&A[0,0],&B[0,0],m,n,k)
	return C

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=2] _dmatmulp(double[:,:] A, double[:,:] B):
	'''
	Matrix multiplication C = A x B
	'''
	cdef int m = A.shape[0], k = A.shape[1], n = B.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] C = np.zeros((m,n),dtype=np.double)
	c_dmatmulp(&C[0,0],&A[0,0],&B[0,0],m,n,k)
	return C

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex64_t,ndim=2] _cmatmulp(np.complex64_t[:,:] A, np.complex64_t[:,:] B):
	'''
	Matrix multiplication C = A x B
	'''
	cdef int m = A.shape[0], k = A.shape[1], n = B.shape[1]
	cdef np.ndarray[np.complex64_t,ndim=2] C = np.zeros((m,n),dtype=np.complex64)
	c_cmatmulp(&C[0,0],&A[0,0],&B[0,0],m,n,k)
	return C

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex128_t,ndim=2] _zmatmulp(np.complex128_t[:,:] A, np.complex128_t[:,:] B):
	'''
	Matrix multiplication C = A x B
	'''
	cdef int m = A.shape[0], k = A.shape[1], n = B.shape[1]
	cdef np.ndarray[np.complex128_t,ndim=2] C = np.zeros((m,n),dtype=np.complex128)
	c_zmatmulp(&C[0,0],&A[0,0],&B[0,0],m,n,k)
	return C

@cr('math.matmulp')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def matmulp(real_full[:,:] A, real_full[:,:] B):
	r'''
	Matrix multiplication in parallel
	C = A x B 

	.. warning::
	A and B are distributed along the processors and C is the same for all of them

	Args:
		A (np.ndarray): Matrix A (M,Q)
		B (np.ndarray): Matrix B (Q,N)
	
	Result:
		np.ndarray: Resulting matrix C (M,N)
	'''
	if real_full is np.complex128_t:
		return _zmatmulp(A,B)
	elif real_full is np.complex64_t:
		return _cmatmulp(A,B)
	elif real_full is double:
		return _dmatmulp(A,B)
	else:
		return _smatmulp(A,B)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=2] _svecmat(float[:] v, float[:,:] A):
	'''
	Vector times a matrix C = v x A
	'''
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] C = np.zeros((m,n),dtype=np.float32)
	memcpy(&C[0,0],&A[0,0],m*n*sizeof(float))
	c_svecmat(&v[0],&C[0,0],m,n)
	return C

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=2] _dvecmat(double[:] v, double[:,:] A):
	'''
	Vector times a matrix C = v x A
	'''
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] C = np.zeros((m,n),dtype=np.double)
	memcpy(&C[0,0],&A[0,0],m*n*sizeof(double))
	c_dvecmat(&v[0],&C[0,0],m,n)
	return C

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex64_t,ndim=2] _cvecmat(np.complex64_t[:] v, np.complex64_t[:,:] A):
	'''
	Vector times a matrix C = v x A
	'''
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.complex64_t,ndim=2] C = np.zeros((m,n),dtype=np.complex64)
	memcpy(&C[0,0],&A[0,0],m*n*sizeof(np.complex64_t))
	c_cvecmat(&v[0],&C[0,0],m,n)
	return C

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex128_t,ndim=2] _zvecmat(np.complex128_t[:] v, np.complex128_t[:,:] A):
	'''
	Vector times a matrix C = v x A
	'''
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.complex128_t,ndim=2] C = np.zeros((m,n),dtype=np.complex128)
	memcpy(&C[0,0],&A[0,0],m*n*sizeof(np.complex128_t))
	c_zvecmat(&v[0],&C[0,0],m,n)
	return C

@cr('math.vecmat')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def vecmat(real_full[:] v, real_full[:,:] A):
	r'''
	Vector times a matrix 
	C = v x A

	Args:
		v (np.ndarray): Vector v (M,)
		A (np.ndarray): Matrix A (M,N)
	
	Result:
		np.ndarray: Resulting matrix C (M,N)
	'''
	if real_full is np.complex128_t:
		return _zvecmat(v,A)
	elif real_full is np.complex64_t:
		return _cvecmat(v,A)
	elif real_full is double:
		return _dvecmat(v,A)
	else:
		return _svecmat(v,A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef void _ssort(float[:] v, int[:] index):
	'''
	Sort a vector
	'''
	cdef int n = v.shape[0]
	c_ssort(&v[0],&index[0],n)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef void _dsort(double[:] v, int[:] index):
	'''
	Sort a vector
	'''
	cdef int n = v.shape[0]
	c_dsort(&v[0],&index[0],n)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef void _csort(np.complex64_t[:] v, int[:] index):
	'''
	Sort a vector
	'''
	cdef int n = v.shape[0]
	c_csort(&v[0],&index[0],n)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef void _zsort(np.complex128_t[:] v, int[:] index):
	'''
	Sort a vector
	'''
	cdef int n = v.shape[0]
	c_zsort(&v[0],&index[0],n)

@cr('math.argsort')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def argsort(real_full[:] v):
	r'''
	Returns the indices that sort a vector

	Args:
		v (np.ndarray): Vector v (M,)
	
	Result:
		np.ndarray: Indices that sort v (M,)
	'''
	cdef int n = v.shape[0]
	cdef np.ndarray[np.int32_t,ndim=1] index = np.zeros((n,),dtype=np.int32)
	if real_full is np.complex128_t:
		return _zsort(v,index)
	elif real_full is np.complex64_t:
		return _csort(v,index)
	elif real_full is double:
		return _dsort(v,index)
	else:
		return _ssort(v,index)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _ceigen(float[:,:] A):
	'''
	Eigenvalues and eigenvectors using Lapack.
		real(n)   are the real eigenvalues.
		imag(n)   are the imaginary eigenvalues.
		vecs(n,n) are the right eigenvectors.
	'''
	cdef int m = A.shape[0], n = A.shape[1], retval
	cdef np.ndarray[np.float32_t,ndim=1]   real = np.zeros((n,),dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=1]   imag = np.zeros((n,),dtype=np.float32)
	cdef np.ndarray[np.complex64_t,ndim=2] vecs = np.zeros((n,n),dtype=np.complex64)
	# Compute eigenvalues and eigenvectors
	retval = c_ceigen(&real[0],&imag[0],&vecs[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing eigenvalues!')
	return real,imag,vecs

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _zeigen(double[:,:] A):
	'''
	Eigenvalues and eigenvectors using Lapack.
		real(n)   are the real eigenvalues.
		imag(n)   are the imaginary eigenvalues.
		vecs(n,n) are the right eigenvectors.
	'''
	cdef int m = A.shape[0], n = A.shape[1], retval
	cdef np.ndarray[np.double_t,ndim=1] real = np.zeros((n,),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=1] imag = np.zeros((n,),dtype=np.double)
	cdef np.ndarray[np.complex128_t,ndim=2] vecs = np.zeros((n,n),dtype=np.complex128)
	# Compute eigenvalues and eigenvectors
	retval = c_zeigen(&real[0],&imag[0],&vecs[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing eigenvalues!')
	return real,imag,vecs

@cr('math.eigen')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def eigen(real[:,:] A):
	r'''
	Eigenvalues and eigenvectors.

	.. warning::
	GPU implementation of this algoritm is still slow and
	thus will be executed purely on CPU level until
	cupy implements linalg.eig

	Args:
		A (np.ndarray): Matrix A (M,N)
	
	Result:
		np.ndarray: the real eigenvalues, real(M)
		np.ndarray: the imaginary eigenvalues, imag(M)
		np.ndarray: the right eigenvectors, vecs(M,M)
	'''
	if real is double:
		return _zeigen(A)
	else:
		return _ceigen(A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _spolar(float[:] rreal, float[:] iimag):
	'''
	Present a complex number in its polar form given its real and imaginary part
	'''
	cdef int i, n = rreal.shape[0]
	cdef np.ndarray[np.float32_t,ndim=1] mod = np.zeros((n,),dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=1] arg = np.zeros((n,),dtype=np.float32)
	for i in range(n):
		mod[i] = sqrt(rreal[i]*rreal[i] + iimag[i]*iimag[i])
		arg[i] = atan2(iimag[i], rreal[i])
	return mod, arg

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _dpolar(double[:] rreal, double[:] iimag):
	'''
	Present a complex number in its polar form given its real and imaginary part
	'''
	cdef int i, n = rreal.shape[0]
	cdef np.ndarray[np.double_t,ndim=1] mod = np.zeros((n,),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=1] arg = np.zeros((n,),dtype=np.double)
	for i in range(n):
		mod[i] = sqrt(rreal[i]*rreal[i] + iimag[i]*iimag[i])
		arg[i] = atan2(iimag[i], rreal[i])
	return mod, arg

@cr('math.polar')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def polar(real[:] rreal, real[:] iimag):
	r'''
	Present a complex number in its polar form given its real and imaginary part

	Args:
		real (np.ndarray): the real component
		imag (np.ndarray): the imaginary component

	Result:
		np.ndarray: the modulus
		np.ndarray: the argument
	'''
	if real is double:
		return _dpolar(rreal,iimag)
	else:
		return _spolar(rreal,iimag)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex64_t,ndim=2] _ccholesky(np.complex64_t[:,:] A):
	'''
	Compute the Lower Cholesky decomposition of matrix A. The C routine modifies directly the matrix!
	'''
	cdef int n = A.shape[0]
	retval = c_ccholesky(&A[0,0], n)
	if not retval == 0: raiseError('Problems computing Cholesky factorization!')
	return np.asarray(A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex128_t,ndim=2] _zcholesky(np.complex128_t[:,:] A):
	'''
	Compute the Lower Cholesky decomposition of matrix A. The C routine modifies directly the matrix!
	'''
	cdef int n = A.shape[0]
	retval = c_zcholesky(&A[0,0], n)
	if not retval == 0: raiseError('Problems computing Cholesky factorization!')
	return np.asarray(A)

@cr('math.cholesky')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def cholesky(real_complex[:,:] A):
	r'''
	Conjugates complex number A

	Args:
		A (np.ndarray): Vector, matrix or number A
	
	Result:
		np.ndarray: Conjugate of A
	'''
	if real_complex is np.complex128_t:
		return _zcholesky(A)
	else:
		return _ccholesky(A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex64_t,ndim=2] _cvandermonde(float[:] rreal, float[:] iimag, int m, int n):
	'''
	Builds a Vandermonde matrix of (m x n) with the real and
	imaginary parts of the eigenvalues
	'''
	cdef np.ndarray[np.complex64_t,ndim=2] vand = np.zeros((m,n),dtype=np.complex64)
	c_cvandermonde(&vand[0,0], &rreal[0], &iimag[0], m, n)
	return np.asarray(vand)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex128_t,ndim=2] _zvandermonde(double[:] rreal, double[:] iimag, int m, int n):
	'''
	Builds a Vandermonde matrix of (m x n) with the real and
	imaginary parts of the eigenvalues
	'''
	cdef np.ndarray[np.complex128_t,ndim=2] vand = np.zeros((m,n),dtype=np.complex128)
	c_zvandermonde(&vand[0,0], &rreal[0], &iimag[0], m, n)
	return np.asarray(vand)

@cr('math.vandermonde')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def vandermonde(real[:] rreal, real[:] iimag, int m, int n):
	r'''
	Builds a Vandermonde matrix of (m x n) with the real and
	imaginary parts of the eigenvalues

	Args:
		real (np.ndarray): the real component
		imag (np.ndarray): the imaginary component
		m (int): number of rows for the Vandermode matrix
		n (int): number of columns for the Vandermode matrix

	Result:
		np.ndarray: the Vandermonde matrix
	'''
	if real is double:
		return _zvandermonde(rreal,iimag,m,n)
	else:
		return _cvandermonde(rreal,iimag,m,n)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex64_t,ndim=2] _cvandermondeTime(float[:] rreal, float[:] iimag, int m, float[:] t):
	'''
	Builds a Vandermonde matrix of (m x n) with the real and
	imaginary parts of the eigenvalues for a certain timesteps
	'''
	cdef int n = t.shape[0]
	cdef np.ndarray[np.complex64_t,ndim=2] vand = np.zeros((m,n),dtype=np.complex64)
	c_cvandermonde_time(&vand[0,0], &rreal[0], &iimag[0], m, n, &t[0])
	return np.asarray(vand)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex128_t,ndim=2] _zvandermondeTime(double[:] rreal, double[:] iimag, int m, double[:] t):
	'''
	Builds a Vandermonde matrix of (m x n) with the real and
	imaginary parts of the eigenvalues for a certain timesteps
	'''
	cdef int n = t.shape[0]
	cdef np.ndarray[np.complex128_t,ndim=2] vand = np.zeros((m,n),dtype=np.complex128)
	c_zvandermonde_time(&vand[0,0], &rreal[0], &iimag[0], m, n, &t[0])
	return np.asarray(vand)

@cr('math.vandermondeTime')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def vandermondeTime(real[:] rreal, real[:] iimag, int m, real[:] t):
	r'''
	Builds a Vandermonde matrix of (m x n) with the real and
	imaginary parts of the eigenvalues


	Args:
		real (np.ndarray): the real component
		imag (np.ndarray): the imaginary component
		m (int): number of rows for the Vandermode matrix
		time (np.ndarray): the time vector

	Result:
		np.ndarray: the Vandermonde matrix
	'''
	if real is double:
		return _zvandermondeTime(rreal,iimag,m,t)
	else:
		return _cvandermondeTime(rreal,iimag,m,t)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=1] _sdiag(float[:,:] A):
	'''
	Returns the diagonal of A (A is a square matrix)
	'''
	cdef int m = A.shape[0]
	cdef int ii
	cdef int jj
	cdef np.ndarray[np.float32_t,ndim=1] B = np.zeros((m,),dtype=np.float32)
	for ii in range(m):
		for jj in range(m):
			B[ii] = A[ii][jj]
	return B

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=1] _ddiag(double[:,:] A):
	'''
	Returns the diagonal of A (A is a square matrix)
	'''
	cdef int m = A.shape[0]
	cdef int ii
	cdef int jj
	cdef np.ndarray[np.double_t,ndim=1] B = np.zeros((m,),dtype=np.double)
	for ii in range(m):
		for jj in range(m):
			B[ii] = A[ii][jj]
	return B

@cr('math.diag')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def diag(real[:,:] A):
	r'''
	If A is a matrix it returns its diagonal, if its a vector it returns
	a diagonal matrix with A in its diagonal

	Args:
		A (np.ndarray): Matrix A (M,N)
	
	Result:
		np.ndarray: Diagonal of A (M,)
	'''
	if real is double:
		return _ddiag(A)
	else:
		return _sdiag(A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex64_t,ndim=2] _cconj(np.complex64_t[:,:] A):
	'''
	Returns the pointwise conjugate of A
	'''
	cdef int m = A.shape[0]
	cdef int n = A.shape[1]
	cdef int ii
	cdef int jj
	cdef np.ndarray[np.complex64_t,ndim=2] B = np.zeros((m,n),dtype=np.complex64)
	for ii in range(m):
		for jj in range(n):
			B[ii, jj] = crealf(A[ii][jj]) - cimagf(A[ii][jj])*I
	return B

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex128_t,ndim=2] _zconj(np.complex128_t[:,:] A):
	'''
	Returns the pointwise conjugate of A
	'''
	cdef int m = A.shape[0]
	cdef int n = A.shape[1]
	cdef int ii
	cdef int jj
	cdef np.ndarray[np.complex128_t,ndim=2] B = np.zeros((m,n),dtype=np.complex128)
	for ii in range(m):
		for jj in range(n):
			B[ii, jj] = creal(A[ii][jj]) - cimag(A[ii][jj])*J
	return B

@cr('math.conj')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def conj(real_complex[:,:] A):
	r'''
	Conjugates complex number A

	Args:
		A (np.ndarray): Vector, matrix or number A
	
	Result:
		np.ndarray: Conjugate of A
	'''
	if real_complex is np.complex128_t:
		return _zconj(A)
	else:
		return _cconj(A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=2] _sinv(float[:,:] A):
	'''
	Returns the inverse of A
	'''
	retval = c_sinv(&A[0,0], A.shape[0], A.shape[1])
	return np.asarray(A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=2] _dinv(double[:,:] A):
	'''
	Returns the inverse of A
	'''
	retval = c_dinv(&A[0,0], A.shape[0], A.shape[1])
	return np.asarray(A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex64_t,ndim=2] _cinv(np.complex64_t[:,:] A):
	'''
	Returns the inverse of A
	'''
	retval = c_cinv(&A[0,0], A.shape[0], A.shape[1])
	return np.asarray(A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.complex128_t,ndim=2] _zinv(np.complex128_t[:,:] A):
	'''
	Returns the inverse of A
	'''
	retval = c_zinv(&A[0,0], A.shape[0], A.shape[1])
	return np.asarray(A)

@cr('math.inv')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def inv(real_full[:,:] A):
	r'''
	Computes the inverse matrix of A

	Args:
		A (np.ndarray): Matrix A (M,N)
	
	Result:
		np.ndarray: Inverse of A (M,N)
	'''
	if real_full is np.complex128_t:
		return _zinv(A)
	elif real_full is np.complex64_t:
		return _cinv(A)
	elif real_full is double:
		return _dinv(A)
	else:
		return _sinv(A)

@cr('math.flip')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def flip(real[:,:] A):
	r'''
	Returns the pointwise conjugate of A

	.. warning::
	This function is not implemented in the
	compiled layer and will raise an error if used

	Args:
		A (np.ndarray): Matrix A (M,N)
	
	Result:
		np.ndarray: Flipped version of A (M,N)
	'''
	raiseError('Function not implemented in Cython!')


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Gappy-POD Module
#
# Last rev: 28/02/2025

from .wrapper import GappyPOD
from .utils import set_random_elements_to_zero, delete_snapshot

del wrapper



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# SPOD general utilities.
#
# Last rev: 17/03/2023
from __future__ import print_function, division

import numpy as np

from ..utils.gpu import cp
from ..          import inp_out as io
from ..utils.cr  import cr_nvtx as cr


@cr('SPOD.extract_modes')
def extract_modes(L,P,ivar,npoints,iblock=1,modes=[],reshape=True):
	'''
	Extract modes for a certain variables
	'''
	p = cp if type(L) is cp.ndarray else np
	nblocks = L.shape[1]
	nvars   = P.shape[0]//(npoints*nblocks)
	# Define modes to extract
	if len(modes) == 0: modes = p.arange(1,P.shape[1]+1,dtype=p.int32)
	# Allocate output array
	out = p.zeros((npoints,len(modes)),L.dtype)
	for m in modes:
		# Extract the block
		Pb         = P[(iblock-1)*nvars*npoints:iblock*nvars*npoints]
		out[:,m-1] = Pb[ivar-1:nvars*npoints:nvars,m-1]
	# Return reshaped output
	return out.reshape((len(modes)*npoints,),order='C') if reshape else out


@cr('SPOD.save')
def save(fname,L,P,f,ptable,nvars=1,pointData=True,mode='w'):
	'''
	Store SPOD variables in serial or parallel
	according to the partition used to compute the SPOD.
	'''
	if type(L) is cp.ndarray:
		io.h5_save_SPOD(fname,L.get(),P.get(),f.get(),ptable,nvars=nvars,pointData=pointData,mode=mode)
	else:
		io.h5_save_SPOD(fname,L,P,f,ptable,nvars=nvars,pointData=pointData,mode=mode)


@cr('SPOD.load')
def load(fname,vars=['L','P','f'],nmod=-1,ptable=None):
	'''
	Load SPOD variables in serial or parallel
	according to the partition used to compute the SPOD.
	'''
	return io.h5_load_SPOD(fname,vars,nmod,ptable)


#!/usr/bin/env python
#
# PYLOM Testsuite
# Build channel dataset for the testsuite
#
# Last revision: 03/08/2021
from __future__ import print_function, division

import numpy as np
import pyAlya, pyLOM


BASEDIR        = '../'
DSETDIR        = './'
CASESTR        = 'channel'
VARLIST        = ['VELOC','PRESS']
START, DT, END = 500000, 500, 505000

# In case of restart, load the previous data
listOfInstants = [ii for ii in range(START,END,DT)]
ni = len(listOfInstants)


## Load pyAlya mesh
mesh = pyAlya.Mesh.read(CASESTR,basedir=BASEDIR,read_commu=False,read_massm=False)
pyAlya.pprint(0,'Run (%d instants)...' % len(listOfInstants),flush=True)


## Create POD dataset
p = pyLOM.PartitionTable.from_pyQvarsi(mesh.partition_table,has_master=True)
m = pyLOM.Mesh.from_pyQvarsi(mesh,ptable=p)


## Build dataset from the instants
time    = np.zeros((len(listOfInstants),),np.double)
X_PRESS = np.zeros((mesh.nnod,ni),dtype=np.double) # POD matrix, VELOC and PRESS
X_VELOX = np.zeros((mesh.nnod,ni),dtype=np.double) # POD matrix, VELOC and PRESS
X_VELOY = np.zeros((mesh.nnod,ni),dtype=np.double) # POD matrix, VELOC and PRESS
X_VELOZ = np.zeros((mesh.nnod,ni),dtype=np.double) # POD matrix, VELOC and PRESS
for ii,instant in enumerate(listOfInstants):
	if ii%100 == 0: pyAlya.pprint(1,'Processing instant %d...'%instant,flush=True)
	# Read data
	field, header = pyAlya.Field.read(CASESTR,VARLIST,instant,mesh.xyz,basedir=BASEDIR)
	# Store time
	time[ii] = header.time
	# Store the POD matrix
	X_PRESS[:,ii] = field['PRESS']
	X_VELOX[:,ii] = field['VELOC'][:,0]
	X_VELOY[:,ii] = field['VELOC'][:,1]
	X_VELOZ[:,ii] = field['VELOC'][:,2]


## Create dataset for pyLOM
d = pyLOM.Dataset(xyz=m.xyz, ptable=p, order=m.pointOrder, point=True,
	# Add the time as the only variable
	vars  = {'time':{'idim':0,'value':time}},
	# Now add all the arrays to be stored in the dataset
	# It is important to convert them as C contiguous arrays
	PRESS = {'ndim':1,'value':X_PRESS},
	VELOX = {'ndim':1,'value':X_VELOX},
	VELOY = {'ndim':1,'value':X_VELOY},
	VELOZ = {'ndim':1,'value':X_VELOZ},
)


## Store dataset
# Nopartition will eliminate the partition of alya and allow the dataset to be run
# with any number of processors but will increase the save time to the disk
# Use with great care!!
m.save('%s.h5'%CASESTR.upper(),nopartition=True) 
d.save('%s.h5'%CASESTR.upper(),nopartition=True) 


pyAlya.cr_info()
pyLOM.cr_info()


#!/usr/bin/env python
#
# POD analysis.
#
# Last revision: 29/10/2021
from __future__ import print_function, division

import mpi4py
mpi4py.rc.recv_mprobe = False

import os, numpy as np
import pyLOM


## Data loading
DATAFILE  = './channel.h5'
VARIABLES = ['PRESS','VELOX','VELOY','VELOZ']

m = pyLOM.Mesh.load(DATAFILE)
d = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table)
X = d.X(*VARIABLES)


## Run POD
PSI,S,V = pyLOM.POD.run(X,remove_mean=False) # PSI are POD modes
if pyLOM.utils.is_rank_or_serial(root=0): 
	fig,_ = pyLOM.POD.plotResidual(S)
	fig.savefig('residuals.png',dpi=300)

# Truncate according to a residual
PSI,S,V = pyLOM.POD.truncate(PSI,S,V,r=3e-2)

# Store results
pyLOM.POD.save('results.h5',PSI,S,V,d.partition_table,nvars=4,pointData=d.point)

# Spatial modes
modes = np.arange(1,100+1,dtype=np.int32)
d.add_field('spatial_modes_P',len(modes),pyLOM.POD.extract_modes(PSI,1,len(d),modes=modes))
d.add_field('spatial_modes_U',len(modes),pyLOM.POD.extract_modes(PSI,2,len(d),modes=modes))
d.add_field('spatial_modes_V',len(modes),pyLOM.POD.extract_modes(PSI,3,len(d),modes=modes))
d.add_field('spatial_modes_W',len(modes),pyLOM.POD.extract_modes(PSI,4,len(d),modes=modes))
pyLOM.io.pv_writer(m,d,'modes',basedir='modes',instants=[0],times=[0.],vars=['spatial_modes_P','spatial_modes_U','spatial_modes_V','spatial_modes_W'],fmt='vtkh5')

# Plot POD modes
if pyLOM.utils.is_rank_or_serial(0):
	# 0 - module, 1,2 - components
	os.makedirs('modes',exist_ok=True)
	figs,_ = pyLOM.POD.plotMode(V[:,:-1],d.time[:-1],modes=[1,2,3,4])
	for ifig,fig in enumerate(figs): fig.savefig('modes/mode_%d.png'%ifig) 


## Reconstruct the flow
X_POD = pyLOM.POD.reconstruct(PSI,S,V)
# Compute RMSE
rmse = pyLOM.math.RMSE(X_POD,X)
pyLOM.pprint(0,'RMSE = %e'%rmse)

# Create a dataset to store the last 500 instants from the simulation
# and the reconstructed flow
idx     = 100
time    = d.get_variable('time')[-idx:]
X_PRESS = X[0:4*d.mesh.npoints:4,-idx:]
X_PRESR = X_POD[0:4*d.mesh.npoints:4,-idx:]
X_VELOC = np.zeros((3*d.mesh.npoints,idx),dtype=np.double)
X_VELOC[0:3*d.mesh.npoints:3,:] = X[1:4*d.mesh.npoints:4,-idx:]
X_VELOC[1:3*d.mesh.npoints:3,:] = X[2:4*d.mesh.npoints:4,-idx:]
X_VELOC[2:3*d.mesh.npoints:3,:] = X[3:4*d.mesh.npoints:4,-idx:]
X_VELOR = np.zeros((3*d.mesh.npoints,idx),dtype=np.double)
X_VELOR[0:3*d.mesh.npoints:3,:] = X_POD[1:4*d.mesh.npoints:4,-idx:]
X_VELOR[1:3*d.mesh.npoints:3,:] = X_POD[2:4*d.mesh.npoints:4,-idx:]
X_VELOR[2:3*d.mesh.npoints:3,:] = X_POD[3:4*d.mesh.npoints:4,-idx:]

d2 = pyLOM.Dataset(xyz=d.xyz, ptable=d.partition_table, vars={'time':{'idim':0,'value':time}}, order=d.ordering, point=d.point)
d2.add_field('PRESS',1,X_PRESS)
d2.add_field('VELOC',3,X_VELOC)
d2.add_field('PRESR',1,X_PRESR)
d2.add_field('VELOR',3,X_VELOR)
pyLOM.io.pv_writer(m,d2,'flow',basedir='flow',instants=np.arange(d2.time.shape[0],dtype=np.int32),times=d2.time,vars=['PRESS','VELOC','PRESR','VELOR'],fmt='vtkh5')


## Show and print timings
pyLOM.cr_info()



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Python interface for POD.
#
# Last rev: 09/07/2021
from __future__ import print_function

from ..vmmath       import vecmat, matmul, temporal_mean, subtract_mean, tsqr_svd, randomized_svd, compute_truncation_residual
from ..utils.cr     import cr_nvtx as cr, cr_start, cr_stop


## POD run method
@cr('POD.run')
def run(X,remove_mean=True, randomized=False, r=1, q=3, seed=-1):
	'''
	Run POD analysis of a matrix X.

	Inputs:
		- X[ndims*nmesh,n_temp_snapshots]: data matrix
		- remove_mean:                     whether or not to remove the mean flow

	Returns:
		- U:  are the POD modes.
		- S:  are the singular values.
		- V:  are the right singular vectors.
	'''
	if remove_mean:
		cr_start('POD.temporal_mean',0)
		# Compute temporal mean
		X_mean = temporal_mean(X)
		# Compute substract temporal mean
		Y = subtract_mean(X,X_mean)
		cr_stop('POD.temporal_mean',0)
	else:
		Y = X.copy()
	# Compute SVD
	cr_start('POD.SVD',0)
	U,S,V = tsqr_svd(Y) if not randomized else randomized_svd(Y,r,q,seed=seed)
	cr_stop('POD.SVD',0)
	# Return
	return U,S,V


## POD truncate method
@cr('POD.truncate')
def truncate(U,S,V,r=1e-8):
	'''
	Truncate POD matrices (U, S, V) given a residual, number of modes or cumulative energy r.

	Inputs:
		- U(m,n)  are the POD modes.
		- S(n)    are the singular values.
		- V(n,n)  are the right singular vectors.
		- r       target residual, number of modes, or cumulative energy threshold.
					* If r >= 1, it is treated as the number of modes.
					* If r < 1 and r > 0 it is treated as the residual target.
					* If r < 1 and r < 0 it is treated as the fraction of cumulative energy to retain.
					Note:  must be in (0,-1] and r = -1 is valid

	Returns:
		- U(m,N)  are the POD modes (truncated at N).
		- S(N)    are the singular values (truncated at N).
		- V(N,n)  are the right singular vectors (truncated at N).
	'''
	# Compute N using S
	N = int(r) if r >= 1 else compute_truncation_residual(S, r)
	
 	# Truncate
	Ur = U[:,:N]
	Sr = S[:N]
	Vr = V[:N,:]
	# Return
	return Ur, Sr, Vr


## POD reconstruct method
@cr('POD.reconstruct')
def reconstruct(U,S,V):
	'''
	Reconstruct the flow given the POD decomposition matrices
	that can be possibly truncated.
	N is the truncated size
	n is the number of snapshots

	Inputs:
		- U(m,N)  are the POD modes.
		- S(N)    are the singular values.
		- V(N,n)  are the right singular vectors.

	Outputs
		- X(m,n)  is the reconstructed flow.
	'''
	# Compute X = U x S x VT
	return matmul(U,vecmat(S,V))



#!/usr/bin/env python
#
# pyLOM, IO
#
# Output for ParaView readable formats
#
# Last rev: 18/09/2024

import os, numpy as np

from .io_vtkh5      import vtkh5_save_mesh, vtkh5_link_mesh, vtkh5_save_field
from ..utils.cr     import cr
from ..utils.errors import raiseError


@cr('Writer.write')
def pv_writer(Mesh,Dataset,casestr,basedir='./',idim=0,instants=[0],times=[0.],vars=[],fmt='vtkh5',mode='w'):
	'''
	Store the data using various formats for ParaView.

	This method differs from save in the fact that save is used 
	to recover the field, write only outputs the data.
	'''
	os.makedirs(basedir,exist_ok=True)
	if fmt.lower() in ['vtkh5','vtkhdf']:
		VTKHDF5Writer(Mesh,Dataset,casestr,basedir,instants,times,vars,idim,mode)
	else:
		raiseError('Format <%s> not implemented!'%fmt)


def VTKHDF5Writer(mesh,dset,casestr,basedir,instants,times,varnames,idim,mode):
	'''
	VTKHDF dataset writer
	'''
	# Create a mesh file
	meshname = os.path.join(basedir,'%s-mesh-vtk.hdf'%(casestr))
	# We set the mode at vtkh5_save_mesh as it will create the file
	vtkh5_save_mesh(meshname,mesh,mesh.partition_table,mode=mode)
	# Loop the instants
	for instant, time in zip(instants,times):
		fieldname = os.path.join(basedir,'%s-%08d-vtk.hdf'%(casestr,instant))
		# Link the mesh on the file
		# We set the mode at vtkh5_link_mesh as it will create the file
		vtkh5_link_mesh(fieldname,'./%s-mesh-vtk.hdf'%(casestr),mode=mode)
		# Write the data on the file
		varDict = {}
		for v in varnames:
			sliced     = tuple([np.s_[:]] + [0 if i != idim else instant for i in range(len(dset[v].shape)-1)])
			varDict[v] = mesh.reshape_var(dset[v][sliced],dset.info(v))
		# Save field must append to the existing file
		vtkh5_save_field(fieldname,instant,time,dset.point,varDict,mesh.partition_table,mode='a')



#!/usr/bin/env python
#
# Example of Kan.
#
# Last revision: 08/01/2024

import os, torch, numpy as np, matplotlib.pyplot as plt
import pyLOM, pyLOM.NN

seed = 19

def load_dataset(fname,inputs_scaler,outputs_scaler):
    '''
    Auxiliary function to load a dataset into a pyLOM
    NN dataset
    '''
    return pyLOM.NN.Dataset.load(
        fname,
        field_names=["CP"],
        add_mesh_coordinates=True,
        variables_names=["AoA", "Mach"],
        inputs_scaler=inputs_scaler,
        outputs_scaler=outputs_scaler,
    )

def print_dset_stats(name,td):
    '''
    Auxiliary function to print the statistics
    of a NN dataset
    '''
    x, y = next(iter(torch.utils.data.DataLoader(td, batch_size=len(td))))
    pyLOM.pprint(0,f'name={name} ({len(td)}), x ({x.shape}) = [{x.min(dim=0)},{x.max(dim=0)}], y ({y.shape}) = [{y.min(dim=0)},{y.max(dim=0)}]')

def true_vs_pred_plot(y_true, y_pred, path):
    """
    Auxiliary function to plot the true vs predicted values
    """
    num_plots = y_true.shape[1]
    plt.figure(figsize=(10, 5 * num_plots))
    for j in range(num_plots):
        plt.subplot(num_plots, 1, j + 1)
        plt.scatter(y_true[:, j], y_pred[:, j], s=1, c="b", alpha=0.5)
        plt.xlabel("True values")
        plt.ylabel("Predicted values")
        plt.title(f"Scatterplot for Component {j+1}")
        plt.grid(True)

    plt.tight_layout()
    plt.savefig(path, dpi=300)

def plot_train_test_loss(train_loss, test_loss, path):
    """
    Auxiliary function to plot the training and test loss
    """
    plt.figure()
    plt.plot(range(1, len(train_loss) + 1), train_loss, label="Training Loss")
    total_epochs = len(test_loss) # test loss is calculated at the end of each epoch
    total_iters = len(train_loss) # train loss is calculated at the end of each iteration/batch
    iters_per_epoch = total_iters // total_epochs
    plt.plot(np.arange(iters_per_epoch, total_iters+1, step=iters_per_epoch), test_loss, label="Test Loss")
    plt.xlabel("Iterations")
    plt.ylabel("Loss")
    plt.title("Training Loss vs Epoch")
    plt.yscale("log")
    plt.legend()
    plt.grid()
    plt.savefig(path, dpi=300)

## Set device
device = pyLOM.NN.select_device("cpu") # Force CPU for this example, if left in blank it will automatically select the device


## Load datasets and set up the results output
BASEDIR = './DATA'
CASESTR = 'NRL7301'
RESUDIR = 'KAN_DLR_airfoil'
pyLOM.NN.create_results_folder(RESUDIR)

input_scaler  = pyLOM.NN.MinMaxScaler()
output_scaler = pyLOM.NN.MinMaxScaler()
td_train = load_dataset(os.path.join(BASEDIR,f'{CASESTR}_TRAIN.h5'),input_scaler,output_scaler)
td_test  = load_dataset(os.path.join(BASEDIR,f'{CASESTR}_TEST.h5'),input_scaler,output_scaler)
td_val   = load_dataset(os.path.join(BASEDIR,f'{CASESTR}_VAL.h5'),input_scaler,output_scaler)

# if we want to split by flight conditions instead of using the provided split, we can do the following
dataset = td_train + td_test + td_val
generator = torch.Generator().manual_seed(seed) # set seed for reproducibility
td_train, td_test, td_val = dataset.get_splits_by_parameters([0.7, 0.15, 0.15], shuffle=True, generator=generator)

print_dset_stats('train',td_train)
print_dset_stats('test', td_test)
print_dset_stats('val',  td_val)

model = pyLOM.NN.KAN(
    input_size=4,
    output_size=1,
    hidden_size=31,
    n_layers=3,
    p_dropouts=0.0,
    layer_type=pyLOM.NN.ChebyshevLayer,
    model_name="kan_example",
    device=device,
    degree=7
)

training_params = {
    "epochs": 20,
    "lr": 1e-5,
    'lr_gamma': 0.95,
    'lr_scheduler_step': 10,
    'batch_size': 8,
    "print_eval_rate": 1,
    "optimizer_class": torch.optim.Adam,
    "lr_kwargs":{
        "gamma": 0.95,
        "step_size": 3 * len(td_train) // 8 # each 3 epochs
    },
    "max_norm_grad": 0.5,
    "save_logs_path":RESUDIR,
}

pipeline = pyLOM.NN.Pipeline(
    train_dataset=td_train,
    test_dataset=td_test,
    valid_dataset=td_val,
    model=model,
    training_params=training_params,
)

training_logs = pipeline.run()


## check saving and loading the model
pipeline.model.save(os.path.join(RESUDIR,"model.pth"))
model = pyLOM.NN.KAN.load(RESUDIR + "/model.pth")
# to predict from a dataset
preds = model.predict(td_test, batch_size=256)
# to predict from a tensor
# preds = model(torch.tensor(dataset_test[:][0], device=model.device)).cpu().detach().numpy()
scaled_preds = output_scaler.inverse_transform(preds)
scaled_y     = output_scaler.inverse_transform(td_test[:][1])

# check that the scaling is correct
pyLOM.pprint(0,scaled_y.min(), scaled_y.max())

evaluator = pyLOM.NN.RegressionEvaluator()
evaluator(scaled_y, scaled_preds)
evaluator.print_metrics()

true_vs_pred_plot(scaled_y, scaled_preds, RESUDIR + '/true_vs_pred.png')
plot_train_test_loss(training_logs['train_loss'], training_logs['test_loss'], RESUDIR + '/train_test_loss.png')

pyLOM.cr_info()
plt.show()


#!/usr/bin/env python
#
# pyAlya, mem.
#
# Memory module for performance profiling.
#
# Last rev: 11/04/2023
from __future__ import print_function, division

import sys, numpy as np, copy, functools, subprocess

from .mpi    import MPI, MPI_RANK, MPI_SIZE, mpi_reduce
from .errors import raiseError


PLATFORM = sys.platform

CHANNEL_DICT = {}

CONVERSION = {
	'kB' : 1.,    # Output is in kB
	'mB' : 0.001,
	'gB' : 1.e-6,
}


class channel(object):
	'''
	This is a channel for the cr counter
	'''
	def __init__(self, name, mmax, mmin, msum, nop, mini):
		self._name = name # Name of the channel
		self._mmax = mmax # Maximum of the channel
		self._mmin = mmin # Minimum of the channel
		self._msum = msum # Total of the channel
		self._nop  = nop  # Number of operations
		self._mini = mini # Initial instant (if == 0 channel is not being take into account)

	def __str__(self):
		return 'name %-20s n %9d min %e max %e avg %e sum %e' % (self.name,self.nop,self.mmin,self.mmax,self.mavg,self.msum)

	def __add__(self, other):
		new = copy.deepcopy(self)
		new._mmax  = max(new._mmax,other._mmax)
		new._mmin  = min(new._mmin,other._mmin)
		new._msum += other._msum
		new._nop  += other._nop 
		return new

	def __iadd__(self, other):
		self._mmax  = max(self._mmax,other._mmax)
		self._mmin  = min(self._mmin,other._mmin)
		self._msum += other._msum
		self._nop  += other._nop 
		return self

	def reset(self):
		'''
		Reset the channel
		'''
		self._mmax = 0.0
		self._mmin = 0.0
		self._msum = 0.0
		self._nop  = 0.0
		self._mini = 0.0

	def restart(self):
		self._mini = 0.0

	def start(self,mini):
		self._mini = mini

	def increase_nop(self):
		self._nop += 1

	def increase_value(self,value):
		self._msum += value

	def set_max(self,value):
		if value > self._mmax or self._nop == 1: self._mmax = value

	def set_min(self,value):
		if value < self._mmin or self._nop == 1: self._mmin = value

	def elapsed(self,value):
		'''
		Negative values are discarded
		'''
		return max(value - self._mini,0.)

	def is_running(self):
		return not self._mini == 0

	@classmethod
	def new(cls,name):
		'''
		Create a new channel
		'''
		return cls(name,0,0,0,0,0)

	@property
	def name(self):
		return self._name
	@property
	def nop(self):
		return self._nop
	@property
	def mmin(self):
		return self._mmin
	@property
	def mmax(self):
		return self._mmax
	@property
	def mavg(self):
		return self._msum/(1.* self._nop)
	@property
	def msum(self):
		return self._msum


def _newch(ch_name):
	'''
	Add a new channel to the list
	'''
	CHANNEL_DICT[ch_name] = channel.new(ch_name)
	return CHANNEL_DICT[ch_name]

def _findch(ch_name):
	'''
	Look for the channel
	'''
	return CHANNEL_DICT[ch_name] if ch_name in CHANNEL_DICT.keys() else None

def _addsuff(ch_name,suff=-1):
	return ch_name if suff <= 0 else '%s%02d' % (ch_name,suff)

def _findch_crash(ch_name):
	'''
	Look for the channel and crash if it does not exist
	'''
	if not ch_name in CHANNEL_DICT.keys():
		raiseError('Channel %s does not exist!' % ch_name)
	return CHANNEL_DICT[ch_name]

def _findch_create(ch_name):
	'''
	Find the channel and if not found create it
	'''
	return CHANNEL_DICT[ch_name] if ch_name in CHANNEL_DICT.keys() else _newch(ch_name)

def _getvalue(units=''):
	'''
	Returns the used memory on an arbitrary instant but fixed.
	'''
	cmd = "cat /proc/meminfo | grep MemFree | cut -d ':' -f 2 | awk '{$1=$1};1' | cut -d ' ' -f 1"
	return int( subprocess.check_output(cmd,shell=True) ) if PLATFORM.lower() == 'linux' else 0

def _reduce_mem(m1,m2,dtype):
	for key in m2.keys():
		if key in m1.keys():
			# Key exists in m1, then simply accumulate
			m1[key] += m2[key]
		else:
			# Key does not exist in cr1, create it new
			m1[key] = m2[key]
	return m1

mem_reduce = MPI.Op.Create(_reduce_mem, commute=True)

def _print_units(c,units):
	f = CONVERSION[units]
	return 'name %-30s n %9d min %e max %e avg %e sum %e' % (c.name,c.nop,f*c.mmin,f*c.mmax,f*c.mavg,f*c.msum)

def _info_serial(units):
	msum_array = np.array([CHANNEL_DICT[key].msum for key in CHANNEL_DICT.keys()])
	name_array = np.array([CHANNEL_DICT[key].name for key in CHANNEL_DICT.keys()])

	ind = np.argsort(msum_array) # sorted indices

	print('\nmem_info, units=%s:'%units,flush=True)
	for ii in ind[::-1]:
		print(_print_units(CHANNEL_DICT[name_array[ii]],units),flush=True)
	print('',flush=True)

def _info_parallel(units):
	CHANNEL_DICT_G = mpi_reduce(CHANNEL_DICT,op=mem_reduce,root=0)

	if MPI_RANK == 0:
		msum_array = np.array([CHANNEL_DICT_G[key].msum for key in CHANNEL_DICT_G.keys()])
		name_array = np.array([CHANNEL_DICT_G[key].name for key in CHANNEL_DICT_G.keys()])	
	
		ind = np.argsort(msum_array) # sorted indices

		print('\nmem_info, units=%s (mpi size: %d):' % (units,MPI_SIZE),flush=True)
		for ii in ind[::-1]:
			print(_print_units(CHANNEL_DICT_G[name_array[ii]],units),flush=True)
		print('',flush=True)


def mem_reset():
	'''
	Delete all channels and start again
	'''
	CHANNEL_DICT = {}

def mem_info(rank=-1,units='kB'):
	'''
	Print information - order by major sum
	'''
	if rank >= 0 and rank == MPI_RANK:
		_info_serial(units)
	else:
		_info_parallel(units)

def mem_start(ch_name,suff):
	'''
	Start the chrono of a channel
	'''
	name_tmp = _addsuff(ch_name,suff)
	channel  = _findch_create(name_tmp)
	if channel.is_running():
		raiseError('Channel %s was already set!'%channel.name)
	channel.start( _getvalue() )

def mem_stop(ch_name,suff):
	'''
	Stop the chrono of a channel
	'''
	end      = _getvalue()
	name_tmp = _addsuff(ch_name,suff)
	channel  = _findch_crash(name_tmp)
	value     = channel.elapsed(end)

	channel.increase_nop()
	channel.set_max(value)
	channel.set_min(value)
	channel.increase_value(value)

	channel.restart()

def mem_value(ch_name,suff):
	'''
	Get the value of a channel that is running; channel keeps running
	'''
	end = _getvalue()
	name_tmp = _addsuff(ch_name,suff)
	channel  = _findch_crash(name_tmp)
	return channel.elapsed(end)

def mem(ch_name,suff=0):
	def decorator(func):
		@functools.wraps(func)
		def wrapper(*args,**kwargs):
			mem_start(ch_name,suff)
			out = func(*args,**kwargs)
			mem_stop(ch_name,suff)
			return out
		return wrapper
	return decorator



#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - averaging.
#
# Last rev: 27/10/2021

cimport cython
cimport numpy as np

import numpy as np

from .cfuncs    cimport real, c_stemporal_mean, c_dtemporal_mean, c_ssubtract_mean, c_dsubtract_mean
from ..utils.cr  import cr


@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=1] _stemporal_mean(float[:,:] X):
	'''
	Temporal mean of matrix X(m,n) where m is the spatial coordinates
	and n is the number of snapshots.
	'''
	cdef int m = X.shape[0], n = X.shape[1]
	cdef np.ndarray[np.float32_t,ndim=1] out = np.zeros((m,),dtype=np.float32)
	# Compute temporal mean
	c_stemporal_mean(&out[0],&X[0,0],m,n)
	# Return
	return out

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=1] _dtemporal_mean(double[:,:] X):
	'''
	Temporal mean of matrix X(m,n) where m is the spatial coordinates
	and n is the number of snapshots.
	'''
	cdef int m = X.shape[0], n = X.shape[1]
	cdef np.ndarray[np.double_t,ndim=1] out = np.zeros((m,),dtype=np.double)
	# Compute temporal mean
	c_dtemporal_mean(&out[0],&X[0,0],m,n)
	# Return
	return out

@cr('math.temporal_mean')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def temporal_mean(real[:,:] X):
	r'''
	Temporal mean of matrix X(m,n) where m is the spatial coordinates
	and n is the number of snapshots.

	Args:
		X (numpy.ndarray): Snapshot matrix (m,n).

	Returns:
		numpy.ndarray: Averaged snapshot matrix (m,).
	'''
	if real is double:
		return _dtemporal_mean(X)
	else:
		return _stemporal_mean(X)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.float32_t,ndim=2] _ssubtract_mean(float[:,:] X, float[:] X_mean):
	'''
	Computes out(m,n) = X(m,n) - X_mean(m) where m is the spatial coordinates
	and n is the number of snapshots.
	'''
	cdef int m = X.shape[0], n = X.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] out = np.zeros((m,n),dtype=np.float32)
	# Compute substract temporal mean
	c_ssubtract_mean(&out[0,0],&X[0,0],&X_mean[0],m,n)
	# Return
	return out

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef np.ndarray[np.double_t,ndim=2] _dsubtract_mean(double[:,:] X, double[:] X_mean):
	'''
	Computes out(m,n) = X(m,n) - X_mean(m) where m is the spatial coordinates
	and n is the number of snapshots.
	'''
	cdef int m = X.shape[0], n = X.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] out = np.zeros((m,n),dtype=np.double)
	# Compute substract temporal mean
	c_dsubtract_mean(&out[0,0],&X[0,0],&X_mean[0],m,n)
	# Return
	return out

@cr('math.subtract_mean')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def subtract_mean(real[:,:] X, real[:] X_mean):
	r'''
	Computes out(m,n) = X(m,n) - X_mean(m) where m is the spatial coordinates
	and n is the number of snapshots.

	Args:
		X (numpy.ndarray): Snapshot matrix (m,n).
		X_mean (numpy.ndarray): Averaged snapshot matrix (m,)

	Returns:
		numpy.ndarray: Snapshot matrix without the average(m,n).
	'''
	if real is double:
		return _dsubtract_mean(X,X_mean)
	else:
		return _ssubtract_mean(X,X_mean)


#!/usr/bin/env python
#
# PYLOM Testsuite
# Build synthetic dataset for the testsuite
#
# Last revision: 24/10/2024
from __future__ import print_function, division

import numpy as np
from scipy.io import loadmat

import pyLOM


## Parameters
OUTFILE     = './DATA/SYNTHETIC.h5'
n_samples   = 1250
noise_level = 0.05


## Generate input features
xyz  = np.random.randn(n_samples,3).astype(np.float32)
data = np.zeros((n_samples,1),np.float32)
xyz[:,2]   = 0.
data[:,0]  = np.sin(xyz[:, 0]) + np.cos(xyz[:, 1]) + 0.5 * xyz[:, 0] ** 2
# Add noise
data[:,0] += noise_level*np.random.randn(n_samples)


## Build time instants
time = np.array([1],np.float32)


## Create partition table and order
ptable = pyLOM.PartitionTable.new(1,n_samples,n_samples)
order  = np.arange(n_samples,dtype=np.int32)
print(ptable)


## Create dataset for pyLOM
# Select a limited number of time instants
d = pyLOM.Dataset(xyz=xyz, ptable=ptable, order=order, point=True,
	# Add the time as the only variable
	vars = {'time':{'idim':0,'value':time}},
	# Now add all the arrays to be stored in the dataset
	# It is important to convert them as C contiguous arrays
	DATA = {'ndim':1,'value':data},
)
print(d)
d.save(OUTFILE) # Store dataset


pyLOM.cr_info()


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# POD Module
#
# Last rev: 09/07/2021

from .wrapper import run, truncate, reconstruct
from .utils   import extract_modes, save, load
from .plots   import plotResidual, plotMode, plotSnapshot


del wrapper, plots


#!/usr/bin/env python
#
# pyLOM, IO
#
# PKL Input Output
#
# Last rev: 30/07/2021
from __future__ import print_function, division

import pickle as pkl

from ..utils.cr     import cr


@cr('pklIO.save')
def pkl_save(fname,obj):
	'''
	Save an object in pkl format
	'''
	f = open(fname,'wb')
	pkl.dump(obj,f)
	f.close()


@cr('pklIO.load')
def pkl_load(fname):
	'''
	Load an object in pkl format
	'''
	f = open(fname,'rb')
	out = pkl.load(f)
	f.close()
	return out
